/*******************************************************
  LCD Modbus Slave (Receiver) - Hello World
  - testing OOP library
  
  description
  - Simple sketch how to use receive data on a Modbus Server (Slave)
  - FC3 Read Multiple Holding register  
  - FC6 Write Single Holding Register   
  - FC16 Write Multiple Holding Register

  structure of this sketch: 
  - hardware definitions for the LCD
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - a LED as output
  - a LCD
  - a MAX485-TTL adapter 

  by noiasca
  2021-12-31 ok
  
 *******************************************************/


/* *******************************************************            
 * Serial Interface
 * **************************************************** */    
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(2, 3);               // RX, TX

// On a Mega you can simply use 
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can use
// a reference to Serial
// HardwareSerial &mySerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;   // leave it on SERIAL_8N1 for SoftSerial
constexpr uint32_t modbus_baud = 19200;      // use only slow speeds with SoftSerial
constexpr byte modbus_enable_pin = 11;       // The GPIO used to control the MAX485 TX pin. Set to 255 if not needed (i.e. a selfsensing module or RS232).

/* **************************************************** *            
 * Modbus 
 * **************************************************** */    

#include <NoiascaModbusServerSimple.h>            // Modbus Slave library (low level RS484 communication)
#include <NoiascaModbusCommon.h>                 // predefined Modbus registers for the library examples
constexpr byte modbus_slave_id = 2;              // Modbus address of this slave

// FC3/FC16 holding registers - not used in this example
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  DI12,              // a digitlRead of a GPIO
  DO13,              // a digitalWrite to GPIO
  ADC0,              // analogRead
  ADC1,              // analogRead
  ADC2,              // analogRead
  ADC3,              // analogRead
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // FC3/FC16 register array

ModbusServer modbusServer(modbus_slave_id, mySerial);

/*
   update the holding register (read the pins) 
   and update the pins according to 
   this function should be called over and over again in loop  
*/
void updateRegisters()
{
  // write new values to GPIO
  digitalWrite(13, holdingRegs[DO13]);
  // update data to be read by the master
  holdingRegs[DI12] = digitalRead(12);
  holdingRegs[ADC0] = analogRead(A0);
  holdingRegs[ADC1] = analogRead(A1);
  holdingRegs[ADC2] = analogRead(A2);
  holdingRegs[ADC3] = analogRead(A3);
}

/*
   if you need to control the transmit pin manually
   this callback will be called before the transmission starts  
*/
void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);
}

/*
   if you need to control the transmit pin manually
   this callback will be called after the transmission has ended  
*/
void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);
}

/*
   this functions can be called,
   when a Modbus Client writes a value with FC6 or several values with FC16
   This example only shows a debug messages to indicate Master activity.   
*/
void cbRegisterWritten(const uint16_t reg, const uint16_t value)
{
  Serial.print(F("write register:")); Serial.print(reg); Serial.print(F(" value:")); Serial.println(value);
}

/*
   this functions can be called,
   when a Modbus Client read values from this Modbus Server (FC3).
   If you don't need the debug messages don't set the callback.
   This example only shows a debug messages to indicate Master activity.
*/
void cbRegisterRead(const uint16_t reg, const uint16_t count)
{
  Serial.print(F("read from start register:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

/* **************************************************** *            
 * Arduino default setup and loop
 * **************************************************** */   
void setup()
{
  Serial.begin(115200);                       // only used for debugging
  Serial.println(F("Modbus LCD Slave OOP"));  // comment this lines when you use the hardware serial for Modbus
  Serial.flush();
  
  mySerial.begin(modbus_baud);              // start the Serial for Modbus
  
  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_slave_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);        // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);
  modbusServer.setPostTransmission(cbPostTransmission);
  modbusServer.setRegisterWritten(cbRegisterWritten);
  modbusServer.setRegisterRead(cbRegisterRead);
  // init your local hardware
  pinMode(13, OUTPUT);
}

void loop()
{
  modbusServer.update();         // call modbus in loop
  updateRegisters();             // handle GPIOs (outputs like relais or LEDs and Inputs)
  // do other things non blocking in loop
}
