/*******************************************************
  Modbus Server (Modbus Slave) - FC4 Input Registers
  FC4 Read Input Registers

  structure of this sketch:
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - several GPIOs configured as inputs to read up to 16 bit
  - a MAX485-TTL adapter

  by noiasca
  2021-12-31 in development

 *******************************************************/

#include <NoiascaModbusServerSimple.h>            // Modbus Slave library (low level RS484 communication)

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbus_slave_id = 2;              // Modbus address of this slave
constexpr byte ledPin = 13;
constexpr byte inputRegisterPin[] {A0, A1, A2, A3}; // GPIO as analog input

constexpr byte modbus_enable_pin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                        // for Softserial
constexpr byte txPin = 3;

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
//#include <SoftwareSerial.h>
//SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
HardwareSerial &mySerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial

/* **************************************************** *
   Modbus
 * **************************************************** */

// Both client and server must know the registers
// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  anyRegister,       // just a dummy register
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbus_slave_id, mySerial);

/*
    Modbus Client callbacks
*/
void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);  // handle the DE/RE pins
}

void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);   // handle the DE/RE pins
}

// callback to evaluate if a FC4 Input Register is valid
bool cbIsValidInputRegister(const uint16_t reg)
{
  switch (reg)
  {
    case 2000 ... 2003 :  // a range of valid input register 
    case 3000 :           // a single valid input register
      return true;
    default :
      return false;
  }
}

// callback to respond to FC4 read: Client has requested an input register
uint16_t cbInputRegisterViewed(const uint16_t reg)
{
  //Serial.print(F("requested input register:")); Serial.println(reg);
  switch (reg)
  {
    case 2000: return analogRead(inputRegisterPin[0]);    // you could read a value on demand
    case 2001: return analogRead(inputRegisterPin[1]);
    case 2002: return analogRead(inputRegisterPin[2]);
    case 2003: return analogRead(inputRegisterPin[3]);
    case 3000: return 3000;                               // you can send back any 2 byte value you need
    default: return 0xEEEE;                               // example for error
  }
}


/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(19200);                 // only used for Debugging
  Serial.println(F("Modbus Server Example ReadCoil"));
  Serial.flush();

  mySerial.begin(modbus_baud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_slave_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);                          // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);          // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);        // if you need a TX enable pin, set a callback to your pin handling

  modbusServer.setInputRegisterViewed(cbInputRegisterViewed);  // set callback for: FC4 Read Input Register
  modbusServer.setIsValidInputRegister(cbIsValidInputRegister);// the server must know the valid FC4 Read Input Registers
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  // do other things non blocking in loop
}
