/*******************************************************
  Modbus Server (Modbus Slave) - FC1 FC5 FC15 Read Write Coils
  FC1 Read Coils
  FC5 Write Single Coil
  FC15 Write Multiple Coils

  structure of this sketch:
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - several GPIOs configured as Output for relays (or LEDs) to simulate "Coils" 
  - a MAX485-TTL adapter

  by noiasca
  2022-01-01 ok

 *******************************************************/

#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbus_slave_id = 2;              // Modbus address of this Modbus Server (slave)
constexpr byte ledPin = 13;
constexpr byte coilPin[] {13, 4, 5, 6};          // GPIO to "coils"
constexpr byte modbus_enable_pin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                        // for Softserial
constexpr byte txPin = 3;

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
//HardwareSerial &mySerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, Modbus prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial. 19200 is OK

/* **************************************************** *
   Modbus
 * **************************************************** */

// Both client and server must know the registers
// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  anyRegister,       // just a dummy register
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbus_slave_id, mySerial);

/*
    Modbus Client callbacks
*/
void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);  // handle the DE/RE pins
}

void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);   // handle the DE/RE pins
}

// Callback for FC1 Read Coils
// the function must return the result for one register = one coil
uint16_t cbReadCoil(const uint16_t reg)
{
  //Serial.print(F("FC1 Read Coils, register:")); Serial.println(reg);
  if (digitalRead(coilPin[reg]) == HIGH)   // you could invert the coils if necessary
    return 1;
  else
    return 0;
}

// Callback used for FC5 Read Coil and FC15 Multiple coils
// the function must be able to set the pin
// value 0x0000 is allways to clear the pin. 
// Any other value should set the pin
// The callback receives 0xFF00 from Modbus FC5 to set a pin. 
// The callback receives 0x0001 from Modbus FC15 to set a pin.
void cbWriteCoil(const uint16_t reg, uint16_t value)
{
  //Serial.print(F("FC5/FC15 Write Coil, register:")); Serial.println(reg);
  if (value == 0x0000)
    digitalWrite(coilPin[reg], LOW);   // you could invert the coils if necessary
  else
    digitalWrite(coilPin[reg], HIGH);
}


/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(19200);                 // only used for Debugging
  Serial.println(F("Modbus Server Example ReadCoil"));
  Serial.flush();

  mySerial.begin(modbus_baud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_slave_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);                          // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);          // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);        // if you need a TX enable pin, set a callback to your pin handling

  modbusServer.setReadCoil(cbReadCoil);       // FC1 Read Coils
  modbusServer.setWriteCoil(cbWriteCoil);     // FC5 Write Single Coil / FC15 Write Multiple Coil
  const size_t noOfCoilPin = sizeof(coilPin) / sizeof(coilPin[0]); 
  modbusServer.setNoOfCoils(noOfCoilPin);            // the server must know how many inputs should be used
  //init local hardware
  for (auto & i : coilPin) pinMode(i, OUTPUT);       // you will need to set your "coils" to output
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  // do other things non blocking in loop
}
