/*******************************************************
  Modbus Server (Modbus Slave) - Example for several Function Codes

  FC   Modbus                             Bit Arduino Example           address in this sketch
  --------------------------------------------------------------------------------------------
  FC1  Read Coils                          1  digitalRead               0
  FC5  Write Single Coil                   1  digitalWrite              0
  FC15 Write Multiple Coils                1  digitalWrite              0
  FC2  Read Discrete Inputs                1  digitalRead               0
  FC4  Read Input Registers               16  analogRead                2000..2003
  FC3  Read Multiple Holding register     16                            1, 1000..1003
  FC6  Write Single Holding Register      16  analogWrite               1, 1000..1003
  FC16 Write Multiple Holding Register    16  analogWrite               1, 1000..1003

  structure of this sketch:
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - several GPIOs configured as Output for relays (or LEDs) to simulate "Coils"
  - a MAX485-TTL adapter

  by noiasca
  2022-01-01 ok

 *******************************************************/

#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbus_slave_id = 2;               // Modbus address of this Modbus Server (slave)
constexpr byte ledPin = 13;
constexpr byte coilPin[] {13, 4};                 // GPIO to "coils"
constexpr byte discreteInputPin[] {7, 8, A2, A3}; // GPIO as discrete Inputs
constexpr byte pwmPin[] {5, 6, 9, 10};            // GPIO for PWM output (on UNO 3, 5, 6, 9, 10, 11)
constexpr byte inputRegisterPin[] {A0, A1};       // GPIO to
constexpr byte modbus_enable_pin = 11;            // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                         // for Softserial
constexpr byte txPin = 3;

constexpr size_t noOfPwmPin = sizeof(pwmPin) / sizeof(pwmPin[0]);
byte pwmValue[noOfPwmPin];                        // store and read back the current pwm value

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
//HardwareSerial &mySerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, Modbus prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial. 19200 is OK

/* **************************************************** *
   Modbus
 * **************************************************** */

// Both client and server must know the registers keeping the register 
// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  programVersion,       //
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbus_slave_id, mySerial);

/*
    Modbus Client callbacks
*/
void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);  // handle the DE/RE pins
}

void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);   // handle the DE/RE pins
}

// Callback for FC1 Read Coils
// the function must return the result for one register = one coil
uint16_t cbReadCoil(const uint16_t reg)
{
  //Serial.print(F("FC1 Read Coils, register:")); Serial.println(reg);
  if (digitalRead(coilPin[reg]) == HIGH)   // you could invert the coils if necessary
    return 1;
  else
    return 0;
}

// Callback for FC2 Read Discrete Inputs
// the function must return the result for one register = one discrete input
uint16_t cbReadDiscreteInput(const uint16_t reg)
{
  //Serial.print(F("FC2 Read Discrete Input, Register:")); Serial.println(reg);
  if (digitalRead(discreteInputPin[reg]) == LOW)
    return 1;
  else
    return 0;
}

// Callback used for FC5 Read Coil and FC15 Multiple coils
// the function must be able to set the pin
// value 0x0000 is allways to clear the pin.
// Any other value should set the pin
// The callback receives 0xFF00 from Modbus FC5 to set a pin.
// The callback receives 0x0001 from Modbus FC15 to set a pin.
void cbWriteCoil(const uint16_t reg, uint16_t value)
{
  //Serial.print(F("FC5/FC15 Write Coil, register:")); Serial.println(reg);
  if (value == 0x0000)
    digitalWrite(coilPin[reg], LOW);   // you could invert the coils if necessary
  else
    digitalWrite(coilPin[reg], HIGH);
}

// callback to evaluate if a FC4 Input Register is valid
bool cbIsValidInputRegister(const uint16_t reg)
{
  switch (reg)
  {
    case 2000 ... 2003 :  // a range of valid input register
    case 2100 :           // a single valid input register
      return true;
    default :
      return false;
  }
}

// callback to respond to FC4 read: Client has requested an input register
uint16_t cbInputRegisterViewed(const uint16_t reg)
{
  //Serial.print(F("requested input register:")); Serial.println(reg);
  switch (reg)
  {
    case 2000: return analogRead(inputRegisterPin[0]);    // you could read a value on demand
    case 2001: return analogRead(inputRegisterPin[1]);
    case 2002: return analogRead(inputRegisterPin[2]);
    case 2003: return analogRead(inputRegisterPin[3]);
    case 2100: return 2100;                               // you can send back any 2 byte value you need
    default: return 0xEEEE;                               // example for error
  }
}

// callback to evaluate if an additional address is valid
bool cbIsValidAddress(const uint16_t reg)
{
  switch (reg)
  {
    case 1000 ... 1003 :  // a range of valid addresses
      return true;
    default :
      return false;
  }
}

// callback to respond to FC6/FC16 write: Client has sent a value to an additional address
void cbRegisterSent(const uint16_t reg, uint16_t value)
{
  //Serial.print(F("sent additional address:")); Serial.println(reg);
  switch (reg)
  {
    case 1000: analogWrite(pwmPin[0], value); pwmValue[0] = value; break;
    case 1001: analogWrite(pwmPin[1], value); pwmValue[1] = value; break;
    case 1002: analogWrite(pwmPin[2], value); pwmValue[2] = value; break;
    case 1003: analogWrite(pwmPin[3], value); pwmValue[3] = value; break;
  }
}

// callback to respond to FC3 read: Client has requested an additional address
uint16_t cbRegisterViewed(const uint16_t reg)
{
  //Serial.print(F("viewed additional address:")); Serial.println(reg);
  switch (reg)
  {
    case 1000: return pwmValue[0]; 
    case 1001: return pwmValue[1];
    case 1002: return pwmValue[2];
    case 1003: return pwmValue[3];
  }
  return 0xEEEE;
}

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  holdingRegs[programVersion] = 1;
  Serial.begin(19200);                 // only used for Debugging
  Serial.println(F("Modbus Server all FC supported by Noiasca Modbus Server Simple"));
  Serial.flush();

  mySerial.begin(modbus_baud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_slave_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);                          // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);          // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);        // if you need a TX enable pin, set a callback to your pin handling

  modbusServer.setReadCoil(cbReadCoil);                        // set callback for: FC1 Read Coils
  modbusServer.setWriteCoil(cbWriteCoil);                      // set callback for: FC5 Write Single Coil / FC15 Write Multiple Coil
  const size_t noOfCoilPin = sizeof(coilPin) / sizeof(coilPin[0]);
  modbusServer.setNoOfCoils(noOfCoilPin);                      // the server must know how many coils should be used

  modbusServer.setReadDiscreteInput(cbReadDiscreteInput);      // set callback for: FC2
  size_t noOfDiscreteInputPin = sizeof(discreteInputPin) / sizeof(discreteInputPin[0]);
  modbusServer.setNoOfDiscreteInputs(noOfDiscreteInputPin);    // the server must know how many discrete inputs should be used

  modbusServer.setInputRegisterViewed(cbInputRegisterViewed);  // set callback for: FC4 Read Input Register
  modbusServer.setIsValidInputRegister(cbIsValidInputRegister);// the server must know the valid FC4 Read Input Registers

  modbusServer.setIsValidAddress(cbIsValidAddress);           // valid additional addresses
  modbusServer.setRegisterSent(cbRegisterSent);               // FC6/FC16 write: Client has sent a value to an additional address
  modbusServer.setRegisterViewed(cbRegisterViewed);           // FC3 read: Client has requested an additional address

  //init local hardware
  for (auto & i : coilPin) pinMode(i, OUTPUT);       // you will need to set "coils" to output
  for (auto & i : pwmPin) pinMode(i, OUTPUT);        // you will need to set "pwm" pins to output
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  // do other things non blocking in loop
}
