/*******************************************************
  LCD Modbus Server (Modbus Slave) Example A - Additional addresses ("holding register")

  only one "holding register" as global variabe
  several other "additional addresses" for FC3/FC6/FC16

  structure of this sketch:
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - 4 LEDs
  - a button
  - a MAX485-TTL adapter

  by noiasca
  2022-01-03

 *******************************************************/

#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbusServerId = 2;              // Modbus address of this slave
constexpr byte ledPin = 13;
constexpr byte pin[] = {4, 5, 6, 7};            // some additional pins
constexpr byte modbusEnablePin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                        // for Softserial
constexpr byte txPin = 3;

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
//#include <SoftwareSerial.h>
//SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
HardwareSerial &mySerial = Serial;

constexpr byte modbusFormat = SERIAL_8N1;        // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbusBaud = 19200;           // use slow speeds with SoftSerial

/* **************************************************** *
   Modbus
 * **************************************************** */

// Both, the client and the server must know the registers
// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  regADC0,           // analog port 0
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbusServerId, mySerial);

void updateRegisters()
{
  static uint32_t previousMillis = 0;
  uint32_t currentMillis = millis();
  if (currentMillis - previousMillis > 500)  // this also helps to "debounce" a button
  {
    previousMillis = currentMillis;
    holdingRegs[regADC0] = analogRead(0);
  }
}

void cbRegisterRead(const uint16_t reg, const uint16_t count)
{
  Serial.print(F("read start register:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

// callback to switch on enabe pin before RS485 TX
void cbPreTransmission()
{
  digitalWrite(modbusEnablePin, HIGH);  // handle the DE/RE pins
}

// callback to switch off enabe pin after RS485 TX
void cbPostTransmission()
{
  digitalWrite(modbusEnablePin, LOW);   // handle the DE/RE pins
}

// callback to evaluate if an additional address is valid
bool cbIsValidAddress(const uint16_t reg)
{
  switch (reg)
  {
    case 1000 ... 1004 :  // a range of valid addresses
    case 5000 ... 5004 :  // another range of valid addresses
    case 8900 :             // a single valid addresses
      return true;
    default :
      return false;
  }
}

// callback to respond to FC3 read: Client has requested an additional address
uint16_t cbRegisterViewed(const uint16_t reg)
{
  Serial.print(F("requested additional address:")); Serial.println(reg);
  switch (reg)
  {
    case 1000: return analogRead(0);  // you could read a value on demand
    case 1001: return analogRead(1);
    case 1002: return analogRead(2);
    case 1003: return analogRead(3);
    case 1004: return analogRead(4);
    case 8900: return 89;           // you can send back any 2 byte value you need
    default: return 0xEEEE;
  }
}

// callback to respond to FC6/FC16 write: Client has sent a value to an additional address
void cbRegisterSent(const uint16_t reg, uint16_t value)
{
  Serial.print(F("sent additional address:")); Serial.println(reg);
  switch (reg)
  {
    case 5000: digitalWrite(pin[0], value); break;
    case 5001: digitalWrite(pin[1], value); break;
    case 5002: digitalWrite(pin[2], value); break;
    case 5003: digitalWrite(pin[3], value); break;
    case 5004: digitalWrite(ledPin, value); break;
  }
}

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(115200);                 // only used for Debugging
  Serial.println(F("Modbus Server Example 0x40"));
  Serial.flush();

  mySerial.begin(modbusBaud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbusBaud, modbusFormat); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbusBaud, modbusFormat, modbusServerId, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbusEnablePin, OUTPUT);                           // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);         // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);       // if you need a TX enable pin, set a callback to your pin handling
  //modbusServer.setRegisterWritten(cbRegisterWritten);       // FC6/FC16 write: Client has written to a server holding register
  modbusServer.setRegisterRead(cbRegisterRead);               // FC3 read: Client has requested server holding registers
  modbusServer.setIsValidAddress(cbIsValidAddress);           // valid additional addresses
  modbusServer.setRegisterSent(cbRegisterSent);               // FC6/FC16 write: Client has sent a value to an additional address
  modbusServer.setRegisterViewed(cbRegisterViewed);           // FC3 read: Client has requested an additional address

  //init other hardware
  pinMode(ledPin, OUTPUT);
  for (auto &i : pin) pinMode(i, OUTPUT);
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  updateRegisters();      // handle GPIOs (outputs like relais or LEDs and Inputs)
  // do other things non blocking in loop
}
