/*******************************************************
  LCD Modbus Server (Modbus Slave) Example E

  This sketch simulates (lot of of) input registers and some holding registers of the Solar Charger
  EPSolar Tracer MPPT Series

  Values can be hardcoded in the sketch.

  You can use the Example "RS485 Halfduplex" from the Modbus Master Libary as client
  (but as of 2022-01-03 there are some bugs in the RS485 Halfduplex Example which might give you
  surprising results).

  hardware
  - a MAX485-TTL adapter

  by noiasca
  2022-01-03

 *******************************************************/

#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbusServerId = 2;              // Modbus address of this slave
constexpr byte ledPin = 13;
constexpr byte modbusEnablePin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                        // for Softserial
constexpr byte txPin = 3;

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
//HardwareSerial &mySerial = Serial;

constexpr byte modbusFormat = SERIAL_8N1;        // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbusBaud = 19200;           // use slow speeds with SoftSerial

/* **************************************************** *
   Modbus
 * **************************************************** */

// FC3/FC6/FC16 holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  programVersion,    //
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

// FC3/FC6/FC16 SettingParameter (read-write) holding registers 0x9000 - 0x9070
// FC1/FC5/FC15 coils (read-write) 2, 5, 6
// FC2 Discrete input (read only) 0x2000, 0x200C

// FC4 input registers
enum class IR       // Input Register
{
  // Rated data (read only) input register (0x3000 - 0x300e)
  // Real-time data (read only) input register
  chargingEquipmentInputVoltage =     0x3100,
  chargingEquipmentInputCurrent =     0x3101,
  chargingEquipmentInputPower =       0x3102,  // 2 byte
  chargingEquipmentInputPowerH =      0x3103,
  chargingEquipmentOutputVoltage =    0x3104,
  chargingEquipmentOutputCurrent =    0x3105,
  chargingEquipmentOutputPower =      0x3106,  // 2 byte
  chargingEquipmentOutputPowerH =     0x3107,
  dischargingEquipmentOutputVoltage = 0x310c,
  dischargingEquipmentOutputCurrent = 0x310d,
  dischargingEquipmentOutputPower =   0x310e,  // 2 byte
  dischargingEquipmentOutputPowerH =  0x310f,
  batteryTemperature =                0x3110,
  temperatureInsideEquipment =        0x3111,
  powerComponentsTemperature =        0x3112,
  batterySoc =                        0x311A,
  remoteBatteryTemperature =          0x311B,
  batterysRealRatedPower =            0x311D,
  // Real-Time status (read only) input registers
  batteryStatus =                     0x3200,
  chargingEquipmentStatus =           0x3201,
  // Statistical parameter (read only) input registers
  maximumInputVoltage =               0x3300,
  minimumInputVoltage =               0x3301,
  maximumBatteryVoltage =             0x3302,
  minimumBatteryVoltage =             0x3303,
  consumedEnergyToday =               0x3304, // 2byte
  consumedEnergyTodayH =              0x3305,
  consumedEnergyMonth =               0x3306, // 2byte
  consumedEnergyMonthH =              0x3307,
  consumedEnergyYear  =               0x3308, // 2byte
  consumedEnergyYearH =               0x3309,
  totalConsumedEnergy =               0x330A, // 2byte
  totalConsumedEnergyH =              0x330B,
  generatedEnergyToday =              0x330C, // 2byte
  generatedEnergyTodayH =             0x330D,
  generatedEnergyMonth =              0x330E, // 2byte
  generatedEnergyMonthH=              0x330F,
  generatedEnergyYear  =              0x3310, // 2byte
  generatedEnergyYearH =              0x3311,
  totalGeneratedEnergy =              0x3312, // 2byte
  totalGeneratedEnergyH =             0x3313,
  carbonDioxideReduction =            0x3314, // 2byte
  carbonDioxideReductionH =           0x3315,
  batteryCurrent =                    0x331B, // 2byte like the manually calculated netto
  batteryCurrentH =                   0x331C,
  batteryTemp =                       0x331D,                       // fast gleich wie solarRemoteBatteryTemperature
  ambientTemp =                       0x331E
};



ModbusServer modbusServer(modbusServerId, mySerial);

void updateRegisters()
{
  static uint32_t previousMillis = 0;
  uint32_t currentMillis = millis();
  if (currentMillis - previousMillis > 500)  // this also helps to "debounce" a button
  {
    previousMillis = currentMillis;
    holdingRegs[0] = analogRead(0);
  }
}

void cbRegisterRead(const uint16_t reg, const uint16_t count)
{
  Serial.print(F("read start register:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

// callback to switch on enabe pin before RS485 TX
void cbPreTransmission()
{
  digitalWrite(modbusEnablePin, HIGH);  // handle the DE/RE pins
}

// callback to switch off enabe pin after RS485 TX
void cbPostTransmission()
{
  digitalWrite(modbusEnablePin, LOW);   // handle the DE/RE pins
}

// callback to evaluate if an additional address is valid
bool cbIsValidAddress(const uint16_t reg)
{
  switch (reg)
  {
    case 0x9000 ... 0x900E :
    case 0x901A ... 0x9021 :
    case 0x903D ... 0x903F :
    case 0x9042 ... 0x904D :
    case 0x9065 ... 0x9070 :
      return true;
    default :
      return false;
  }
}

// callback to respond to FC3 read: Client has requested an additional address
uint16_t cbRegisterViewed(const uint16_t reg)
{
  Serial.print(F("requested additional address:")); Serial.print(reg);
  uint16_t result = 0xEEEE; // assume error;
  switch (reg)
  {
    case 0x9000 ... 0x900E : result = random (100, 200); break;
    case 0x901A ... 0x9021 : result = random (200, 400); break;
    case 0x903D ... 0x903F : result = random (300, 500); break;
    case 0x9042 ... 0x904D : result = random (600, 700); break;
    case 0x9065 ... 0x9070 : result = random (800, 900); break;
    default:
      Serial.print("/0x");
      Serial.print(reg, HEX);
      Serial.print(F(" is unknown"));
  }
  Serial.println();
  return result;
}

// callback to respond to FC6/FC16 write: Client has sent a value to an additional address
void cbRegisterSent(const uint16_t reg, uint16_t value)
{
  Serial.print(F("sent additional address:")); Serial.println(reg);
  (void)value;
}

// callback to evaluate if a FC4 Input Register is valid
bool cbIsValidInputRegister(const uint16_t reg)
{
  switch (reg)
  {
    case (uint16_t)IR::chargingEquipmentInputVoltage :
    case (uint16_t)IR::chargingEquipmentInputCurrent :
    case (uint16_t)IR::chargingEquipmentInputPower :
    case (uint16_t)IR::chargingEquipmentInputPowerH:
    case (uint16_t)IR::chargingEquipmentOutputVoltage :
    case (uint16_t)IR::chargingEquipmentOutputCurrent :
    case (uint16_t)IR::chargingEquipmentOutputPower :
    case (uint16_t)IR::chargingEquipmentOutputPowerH:
    case (uint16_t)IR::dischargingEquipmentOutputVoltage :
    case (uint16_t)IR::dischargingEquipmentOutputCurrent :
    case (uint16_t)IR::dischargingEquipmentOutputPower :
    case (uint16_t)IR::dischargingEquipmentOutputPowerH:
    case (uint16_t)IR::batteryTemperature :
    case (uint16_t)IR::temperatureInsideEquipment :
    case (uint16_t)IR::powerComponentsTemperature :
    case (uint16_t)IR::batterySoc:
    case (uint16_t)IR::remoteBatteryTemperature :
    case (uint16_t)IR::batterysRealRatedPower :
    case (uint16_t)IR::batteryStatus:
    case (uint16_t)IR::chargingEquipmentStatus :
    case (uint16_t)IR::maximumInputVoltage:
    case (uint16_t)IR::minimumInputVoltage :
    case (uint16_t)IR::maximumBatteryVoltage :
    case (uint16_t)IR::minimumBatteryVoltage :
    case (uint16_t)IR::consumedEnergyToday :
    case (uint16_t)IR::consumedEnergyTodayH:
    case (uint16_t)IR::consumedEnergyMonth :
    case (uint16_t)IR::consumedEnergyMonthH:
    case (uint16_t)IR::consumedEnergyYear  :
    case (uint16_t)IR::consumedEnergyYearH:
    case (uint16_t)IR::totalConsumedEnergy :
    case (uint16_t)IR::totalConsumedEnergyH:
    case (uint16_t)IR::generatedEnergyToday :
    case (uint16_t)IR::generatedEnergyTodayH :
    case (uint16_t)IR::generatedEnergyMonth :
    case (uint16_t)IR::generatedEnergyMonthH :
    case (uint16_t)IR::generatedEnergyYear  :
    case (uint16_t)IR::generatedEnergyYearH :
    case (uint16_t)IR::totalGeneratedEnergy :
    case (uint16_t)IR::totalGeneratedEnergyH :
    case (uint16_t)IR::carbonDioxideReduction :
    case (uint16_t)IR::carbonDioxideReductionH :
    case (uint16_t)IR::batteryCurrent :
    case (uint16_t)IR::batteryCurrentH :
    case (uint16_t)IR::batteryTemp :
    case (uint16_t)IR::ambientTemp :
      return true;
    default :
      return false;
  }
}

// callback to respond to FC4 read: Client has requested an input register
uint16_t cbInputRegisterViewed(const uint16_t reg)
{
  Serial.print(F("input register viewed:")); Serial.println(reg);
  switch (reg)
  {
    case (uint16_t)IR::chargingEquipmentInputVoltage      : return 1280;
    case (uint16_t)IR::chargingEquipmentInputCurrent      : return 123;
    case (uint16_t)IR::chargingEquipmentInputPower        : return 4567;
    case (uint16_t)IR::chargingEquipmentInputPowerH       : return 0;
    case (uint16_t)IR::chargingEquipmentOutputVoltage     : return 1299;
    case (uint16_t)IR::chargingEquipmentOutputCurrent     : return 45;
    case (uint16_t)IR::chargingEquipmentOutputPower       : return 678;
    case (uint16_t)IR::chargingEquipmentOutputPowerH      : return 0;
    case (uint16_t)IR::dischargingEquipmentOutputVoltage  : return 1270;
    case (uint16_t)IR::dischargingEquipmentOutputCurrent  : return 234;
    case (uint16_t)IR::dischargingEquipmentOutputPower    : return 123;
    case (uint16_t)IR::dischargingEquipmentOutputPowerH   : return 1;
    case (uint16_t)IR::batteryTemperature                 : return 2010;
    case (uint16_t)IR::temperatureInsideEquipment         : return 2560;
    case (uint16_t)IR::powerComponentsTemperature         : return 2570;
    case (uint16_t)IR::batterySoc                         : return 98;  // in %
    case (uint16_t)IR::remoteBatteryTemperature           : return 1234;
    case (uint16_t)IR::batterysRealRatedPower             : return 200;
    //                                                               W-------TTTTVVVV
    case (uint16_t)IR::batteryStatus                      : return 0b1000000000100010;
    //                                                               IIMAIOLM---PCCFR
    case (uint16_t)IR::chargingEquipmentStatus            : return 0b1011111100011111;
    case (uint16_t)IR::maximumInputVoltage                : return 1370;
    case (uint16_t)IR::minimumInputVoltage                : return 1150;
    case (uint16_t)IR::maximumBatteryVoltage              : return 1480;
    case (uint16_t)IR::minimumBatteryVoltage              : return 1230;
    case (uint16_t)IR::consumedEnergyToday                : return 1234;
    case (uint16_t)IR::consumedEnergyTodayH               : return 0;
    case (uint16_t)IR::consumedEnergyMonth                : return 5678;
    case (uint16_t)IR::consumedEnergyMonthH               : return 0;
    case (uint16_t)IR::consumedEnergyYear                 : return 9012;
    case (uint16_t)IR::consumedEnergyYearH                : return 0;
    case (uint16_t)IR::totalConsumedEnergy                : return 3456;
    case (uint16_t)IR::totalConsumedEnergyH               : return 0;
    case (uint16_t)IR::generatedEnergyToday               : return 7890;
    case (uint16_t)IR::generatedEnergyTodayH              : return 0;
    case (uint16_t)IR::generatedEnergyMonth               : return 1280;
    case (uint16_t)IR::generatedEnergyMonthH              : return 0;
    case (uint16_t)IR::generatedEnergyYear                : return 1280;
    case (uint16_t)IR::generatedEnergyYearH               : return 0;
    case (uint16_t)IR::totalGeneratedEnergy               : return 1280;
    case (uint16_t)IR::totalGeneratedEnergyH              : return 0;
    case (uint16_t)IR::carbonDioxideReduction             : return 1280;
    case (uint16_t)IR::carbonDioxideReductionH            : return 0;
    case (uint16_t)IR::batteryCurrent                     : return 1280;
    case (uint16_t)IR::batteryCurrentH                    : return 1280;
    case (uint16_t)IR::batteryTemp                        : return 2047;
    case (uint16_t)IR::ambientTemp                        : return 3247;
    default :
      return 0xEEEE;
  }
}

// Callback for FC1 Read Coils
// The EPSolar has 3 coils
// 2 Manual control the load
// 5 Enable lod test mode
// 6 Force the load on/off
uint16_t cbReadCoil(const uint16_t reg)
{
  Serial.print(F("FC1 Read Coils, register:")); Serial.println(reg);
  if (reg == 2) 
  {
  if (digitalRead(ledPin) == HIGH)   // you could invert the coils if necessary
    return 1;
  else
    return 0;
  }
  
  return 0;
}

// Callback used for FC5 Read Coil and FC15 Multiple coils
void cbWriteCoil(const uint16_t reg, uint16_t value)
{
  Serial.print(F("FC5/FC15 Write Coil, register:")); Serial.println(reg);
  if (reg == 2 && value == 0x0000)
    digitalWrite(ledPin, LOW);   // you could invert the coils if necessary
  else
    digitalWrite(ledPin, HIGH);
}

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(19200);                 // only used for Debugging
  Serial.println(F("Modbus Server Example Simulate EPSolar"));
  Serial.flush();
  holdingRegs[programVersion] = 1;

  mySerial.begin(modbusBaud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbusBaud, modbusFormat); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbusBaud, modbusFormat, modbusServerId, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbusEnablePin, OUTPUT);                           // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);         // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);       // if you need a TX enable pin, set a callback to your pin handling

  modbusServer.setIsValidAddress(cbIsValidAddress);           // valid additional addresses
  modbusServer.setRegisterSent(cbRegisterSent);               // FC6/FC16 write: Client has sent a value to an additional address
  modbusServer.setRegisterViewed(cbRegisterViewed);           // FC3 read: Client has requested an additional address

  modbusServer.setInputRegisterViewed(cbInputRegisterViewed);  // set callback for: FC4 Read Input Register
  modbusServer.setIsValidInputRegister(cbIsValidInputRegister);// the server must know the valid FC4 Read Input Registers

  modbusServer.setReadCoil(cbReadCoil);       // FC1 Read Coils
  modbusServer.setWriteCoil(cbWriteCoil);     // FC5 Write Single Coil / FC15 Write Multiple Coil
  modbusServer.setNoOfCoils(7);               // Coils 2, 5, 6 are useable

  //init other hardware
  pinMode(ledPin, OUTPUT);
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  // do other things non blocking in loop
}
