/*******************************************************
  LCD Modbus Server (Slave) - additional Addresses
  - additional addresses as valid registers
  
  structure of this sketch:
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library

  hardware
  - a servo
  - a LED
  - a button
  - a MAX485-TTL adapter

  by noiasca
  2022-01-01

 *******************************************************/

#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm
#include <Servo.h>

/* **************************************************** *
   settings, variables, objects ...
 * **************************************************** */

constexpr byte modbus_slave_id = 2;              // Modbus address of this slave
constexpr byte ledPin = 13;
constexpr byte servoPin = 12;
constexpr byte buttonPin = A1;                   // GPIO to GND
constexpr byte modbus_enable_pin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;                        // for Softserial
constexpr byte txPin = 3;

Servo servo;

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(rxPin, txPin);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
// HardwareSerial &mySerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial

/* **************************************************** *
   Modbus
 * **************************************************** */

// Both client and server must know the registers
// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  regButton,         // read a button
  regADC0,           // analog port 0
  regLED,            // set a LED
  regServo,          // set a Servo will need two bytes
  HOLDING_REGS_SIZE  // leave this one
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbus_slave_id, mySerial);

void updateRegisters()
{
  static uint32_t previousMillis = 0;
  uint32_t currentMillis = millis();
  if (currentMillis - previousMillis > 50)  // this also helps to "debounce" a button
  {
    previousMillis = currentMillis;
    if (digitalRead(buttonPin) == LOW)
      holdingRegs[regButton] = 1;
    else
      holdingRegs[regButton] = 0;
    holdingRegs[regADC0] = analogRead(0);
  }
}

void updateLocalHardware()
{
  // switch the LED
  if (holdingRegs[regLED] > 0) digitalWrite(ledPin, HIGH); else digitalWrite(ledPin, LOW);
  // move the servo
  int value = holdingRegs[regServo];
  servo.write(value);
}

/*
    Modbus Client callbacks
*/
void cbRegisterWritten(const uint16_t reg, const uint16_t value) // optional
{
  Serial.print(F("write register:")); Serial.print(reg); Serial.print(F(" value:")); Serial.println(value);
  // after request we assume that some of the
  // holding registers have new values
  // hence we control hardware:
  updateLocalHardware();
}

void cbRegisterRead(const uint16_t reg, const uint16_t count)
{
  Serial.print(F("read start register:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);  // handle the DE/RE pins
}

void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);   // handle the DE/RE pins
}

// demo
bool cbIsValidAddress(const uint16_t reg)
{
  switch (reg)
  {
    case 1000 ... 1003 :
    case 1010 :
    case 1020 :
    case 1030 :
      return true;
    default :
      return false;
  }
}

uint16_t cbRegisterViewed(const uint16_t reg)
{
  Serial.print(F("requested additional register adress:")); Serial.println(reg);
  switch (reg)
  {
    case 1000: return 0;
    case 1001: return 11;
    case 1002: return 22;
    case 1003: return 33;
    case 1010: return 10;
    case 1020: return 20;
    case 1030: return 30;

    default: return 0xEEEE;
  }
}

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(115200);                 // only used for Debugging
  Serial.println(F("Modbus Server Example 0x50"));

  mySerial.begin(modbus_baud);          // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_slave_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);                          // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);          // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);        // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setRegisterWritten(cbRegisterWritten);
  modbusServer.setRegisterRead(cbRegisterRead);

  modbusServer.setIsValidAddress(cbIsValidAddress); // demo
  modbusServer.setRegisterViewed(cbRegisterViewed); // demo
  //init other hardware
  pinMode(ledPin, OUTPUT);
  pinMode(buttonPin, INPUT_PULLUP);
  servo.attach(servoPin);
}

void loop()
{
  modbusServer.update();   // call modbus in loop
  updateRegisters();      // handle GPIOs (outputs like relais or LEDs and Inputs)
  // do other things non blocking in loop
}
