/*******************************************************
  LCD Modbus Server - Modbus Slave - (Receiver)

  show splash screen for n seconds after boot
  handle remaining IOs

  structure of this sketch:
  - hardware definitions for the LCD
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library
  - additional Hardware definitions for LEDs or Buttons on your Modbus Server

  hardware
  - a LCD connected to the Arduino
  - a MAX485-TTL adapter
  - optional: some buttons connected to a resistor network to get key input
  - optional: indicator LEDs

  by noiasca
  2021-12-21  function renaming
  2021-11-03  OK

 *******************************************************/

/* ******************************************************
   LCD
 * *************************************************** */

#include <NoiascaLiquidCrystal.h>      // download library from https://werner.rothschopf.net/202009_arduino_liquid_crystal_intro.htm
#include <NoiascaHW/lcd_4bit.h>        // parallel interface, 4bit

constexpr byte cols = 16;                  // columns/characters per row
constexpr byte rows = 2;                   // how many rows
constexpr byte rs = 8;                     // RS pin
constexpr byte en = 9;                     // EN pin
constexpr byte d4 = 4;                     // data pin
constexpr byte d5 = 5;
constexpr byte d6 = 6;
constexpr byte d7 = 7;
constexpr byte bl = 10;                    // backlight pin
constexpr t_backlightPol blType = FAULTY;  // set to either POSITIVE, NEGATIVE, FAULTY (for some keypad shields)

//Create the LCD object:
LiquidCrystal_4bit lcd(rs, en, d4, d5, d6, d7, bl, blType, cols, rows);                   /// Standard UTF-8 converter

/* *******************************************************
   Serial Interface
 * **************************************************** */
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial
constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr byte modbus_enable_pin = 11;           // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter
constexpr byte rxPin = 2;
constexpr byte txPin = 3;

// if you don't have enough HW Serial (e.g. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial mySerial(rxPin, txPin);

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &mySerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
// HardwareSerial &mySerial = Serial;

/* **************************************************** *
   Modbus
 * **************************************************** */
#include <NoiascaModbusServerSimple.h>           // Modbus Server Simple library https://werner.rothschopf.net/microcontroller/202112_arduino_modbus_server.htm
#include <NoiascaModbusCommon.h>                 // predefined Modbus registers for the library examples
#include <NoiascaModbusServerLCDReceiver.h>

constexpr byte modbus_server_id = 2;              // Modbus address of this Server

// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  DI12,              // SI:
  DO13,              // SO: a digitalWrite to GPIO
  ADC0,              // SI: analogRead
  ADC1,              // SI: analogRead
  ADC2,              // SI: analogRead
  ADC3,              // SI: analogRead
  HOLDING_REGS_SIZE  // leave this one
  // total number of registers for function 3 and 16 share the same register array
  // i.e. the same address space
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

ModbusServer modbusServer(modbus_server_id, mySerial);

/* **************************************************** *
   Modbus
 * **************************************************** */
#include "LCD.h"  // contains some helpers and functions for 

void updateRegisters()
{
  // write new values to GPIO
  digitalWrite(13, holdingRegs[DO13]);
  // update data to be read by the master
  holdingRegs[DI12] = digitalRead(12);
  analogRead(A0);
  holdingRegs[ADC0] = analogRead(A0);
  holdingRegs[ADC1] = analogRead(A1);
  holdingRegs[ADC2] = analogRead(A2);
  holdingRegs[ADC3] = analogRead(A3);
}

uint32_t previousMasterMessageUser = 42; // MISSING gehört noch ersetzt ist nur hier zum kompilieren...

/*
    Modbus Master callbacks
*/
void cbPreTransmission()
{
  digitalWrite(modbus_enable_pin, HIGH);
}

void cbPostTransmission()
{
  digitalWrite(modbus_enable_pin, LOW);
}

void cbOnClientActivity()
{
  previousMasterMessageUser = millis();
  indicatorMasterTimeout.off();
  //indicatorMasterOnline.on();
}

void statusTimer()
{
  // check if master is offline
  if (millis() - previousMasterMessageUser > 10000)
  {
    indicatorMasterTimeout.on();
    //indicatorMasterTimeout.on();
  }
  else
  {
    //indicatorMasterTimeout.off();
    //indicatorMasterTimeout.on();
  }
  //heartbeat
}

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(115200);                 // only used for Debugging
  Serial.println(F("Modbus LCD Server GPIO OOP"));

  mySerial.begin(modbus_baud);                  // start the Serial used for Modbus. Remember: Softserial doesn't support modbus_format
  //mySerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbusServer.begin(&mySerial, modbus_baud, modbus_format, modbus_server_id, HOLDING_REGS_SIZE, holdingRegs);
  pinMode(modbus_enable_pin, OUTPUT);                     // if you need a TX enable pin, set it to Output
  modbusServer.setPreTransmission(cbPreTransmission);      // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setPostTransmission(cbPostTransmission);    // if you need a TX enable pin, set a callback to your pin handling
  modbusServer.setIsValidAddress(mslrIsValidAddress);      // mandatory
  modbusServer.setRegisterSent(mslrRegisterSent);              // mandatory: inform the modbus Server about the destination of LCD registers
  modbusServer.setStreamEnd(mslrStreamEnd);                    // mandatory: when a FC16/stream has ended (i.e. for createChar)
  modbusServer.setRegisterWritten(mslrRegisterWritten);        // optional: a simple debug print for write registers (FC3)
  modbusServer.setRegisterRead(mslrRegisterRead);
  modbusServer.setOnClientActivity(cbOnClientActivity);
   
  lcd.begin();                         // initialize the display
  lcd.backlight();
  lcd.setCursor(0, 0);

  if (splashscreenTimeout > 0) showSplashscreen();
  delay(1000);
  if (splashscreenTimeout > 0) showConfig();
  
  //init other hardware
  pinMode(13, OUTPUT);
}

void loop()
{
  modbusServer.update();          // call modbus in loop
  updateRegisters();             // handle GPIOs (outputs like relais or LEDs and Inputs)

  lcdTimer();                    // call time management for the LCD
  statusTimer();

  //indicatorMasterOnline.update();
  indicatorMasterTimeout.update();

  // do other things non blocking in loop
}
