/* ******************************************************
  LCD Modbus Master (Sender) - Simple

  description:
  - The Master will send it's uptime every 3 seconds to the Slave.
    Additionally the Master will retrieve data from the Slave and print them on Serial.
    Finally a port will be toggled on the Slave.
  - use this Modbus Master with a Slave device which responds to Registers 0, 1, and 1001

  structure of this sketch:
  - definition of the Modbus library
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - LCD Modbus library to enable the LCD API 1.0 on the Master Sketch

  hardware
  - a MAX485-TTL adapter

  comment
  - there are two warnings from the Modbus Master Library (not my fault)

  by noiasca
  2021-12-19 compiles
  2021-11-14 OK

 ****************************************************** */

#include <ModbusMaster.h>
#include <NoiascaModbusCommon.h>                 // predefined Modbus registers for the library examples

/*
  We're using a MAX485-compatible RS485 Transceiver.
  Rx/Tx is hooked up to the hardware serial port at 'Serial'.
  The Data Enable and Receiver Enable pins are hooked up as follows:
*/

constexpr byte modbus_format = SERIAL_8N1;       // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;          // use slow speeds with SoftSerial
constexpr byte modbus_enable_pin = 5;            // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter

uint32_t previousTx = 0;
uint16_t restTx = 15000; // rest time between transmissions - microseconds

// instantiate ModbusMaster object
ModbusMaster node;
constexpr byte modbus_slave_address = 2;         // Address of the slave with LCD
constexpr byte wait = 10;                         // 3ms are ok for 19200

/*
    Modbus Master callbacks
*/
void preTransmission()
{
  while (micros() - previousTx < restTx)
  {
    yield(); // wait some time and do background tasks
  }
  digitalWrite(modbus_enable_pin, HIGH);
}

void postTransmission()
{
  previousTx = micros();   // remember last timestamp
  digitalWrite(modbus_enable_pin, LOW);
}

// This LCD interface connects to Modbus via Serial
#include <lcd_Modbus_BufferPrint.h>              // this LCD class writes to Modbus interface but can be used like LCD 
LiquidCrystal_Modbus lcd(node);                  // you must hand over the Modbus Object (node) to the LCD sender

/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(115200);
  Serial.println(F("Modbus LCD Master simple"));

  // Modbus communication baud
  Serial3.begin(modbus_baud);

  // Init enable pins for modbus master library
  pinMode(modbus_enable_pin, OUTPUT);
  digitalWrite(modbus_enable_pin, LOW);

  // Modbus slave and Serial modification
  node.begin(modbus_slave_address, Serial3);

  // Callbacks allow us to configure the RS485 transceiver correctly
  node.preTransmission(preTransmission);
  node.postTransmission(postTransmission);

  lcd.begin();
  lcd.setDelayPostWrite(wait);         // default 5 for 9600 Baud - set some delay time
  lcd.clear();
  lcd.backlight();

  lcd.print(F("* Master Start * "));   // show a splash screen
}

void loop()
{
  static bool state = true;
  state = !state;
  //lcd.setBacklight(state);           // switch on/off backlight
  lcd.setCursor(0, 1);
  lcd.write('X' + state);              // simple write
  lcd.print(" abc ");                  // print
  lcd.setCursor(6, 1);
  lcd.print(millis() / 1000);          // show master runtime on slave LCD
  uint8_t result;
  result = node.writeSingleRegister(1, state);   // write to slave FunctionCode 6
  delay(wait);                         // give the slave time to settle
  Serial.print("result A="); Serial.println(result, HEX);
  // read from slave registers
  uint8_t startRegister = 2;
  uint8_t noOfRegisters = 1; // was 3

  result = node.readHoldingRegisters(startRegister, noOfRegisters);  // reading at address 0 and read 2 registers with Function Code 3
  delay(wait);                         // give the slave time to settle
  if (result == node.ku8MBSuccess)
  {
    for (uint8_t i = 0; i < noOfRegisters; i++)
    {
      Serial.print(node.getResponseBuffer(i));
      Serial.print("\t");
    }
    Serial.println();
  }
  else
  {
    Serial.print("result B="); Serial.println(result, HEX);
  }

  delay(3000);
}
