/*******************************************************
  LCD Modbus Slave (Receiver)

  show splash screen for n seconds after boot
  handle remaining IOs

  structure of this sketch:
  - hardware definitions for the LCD
  - definitions for the Serial interface - either Software or Hardware and if necessary the TX enable pin
  - definition of the Modbus library
  - additional Hardware definitions for LEDs or Buttons on your Modbus Slave

  hardware
  - a LCD connected to the Arduino
  - a MAX485-TTL adapter
  - optional: some buttons connected to a resistor network to get key input
  - optional: indicator LEDs

  by noiasca
  2021-10-28  OK

 *******************************************************/

/* ******************************************************
   LCD
 * *************************************************** */

#include <NoiascaLiquidCrystal.h>      // download library from https://werner.rothschopf.net/202009_arduino_liquid_crystal_intro.htm
#include <NoiascaHW/lcd_4bit.h>        // parallel interface, 4bit

constexpr byte cols = 16;                  // columns/characters per row
constexpr byte rows = 2;                   // how many rows
constexpr byte rs = 8;                     // RS pin
constexpr byte en = 9;                     // EN pin
constexpr byte d4 = 4;                     // data pin
constexpr byte d5 = 5;
constexpr byte d6 = 6;
constexpr byte d7 = 7;
constexpr byte bl = 10;                    // backlight pin
constexpr t_backlightPol blType = FAULTY;  // set to either POSITIVE, NEGATIVE, FAULTY (for some keypad shields)

//Create the LCD object:
LiquidCrystal_4bit lcd(rs, en, d4, d5, d6, d7, bl, blType, cols, rows);                   /// Standard UTF-8 converter

//                     123456789A123456789B  // a screen shown after startup
#define SPLASHSCREEN0 "** Modbus LCD **"
#define SPLASHSCREEN1 "     V1.0.0"
#define SPLASHSCREEN2 ""
#define SPLASHSCREEN3 ""

constexpr byte splashscreenTimeout = 10;     // in seconds. 0 = no splash screen

// additional Hardware, for example some buttons

enum {btnRIGHT, btnUP, btnDOWN, btnLEFT, btnSELECT, btnNONE};

// define some values used by the panel and buttons
byte lcd_key    = 0;  // the pressed button
int adc_key_in  = 0;  // the readed ADC value
int adc_old     = 0;  // the previous ADC value

// read the buttons
byte readLCDbuttons()
{
  adc_key_in = analogRead(0);      // read the value from the sensor
  //Serial.println(adc_key_in);    // check the readed adc value 
  // my buttons are centered at these values: 0,  99, 254, 407, 639
  // add approx 50 to those values and check if acutal value is close
  if (adc_key_in > 1000) return btnNONE;    // 1st option will be the most likely result
  if (adc_key_in < 50)   return btnRIGHT;
  if (adc_key_in < 195)  return btnUP;
  if (adc_key_in < 380)  return btnDOWN;
  if (adc_key_in < 555)  return btnLEFT;
  if (adc_key_in < 790)  return btnSELECT;
  return btnNONE;                           // when all others fail, return this...
}

//const byte arrowUp[8] =   {0b00000, 0b00100, 0b01110, 0b10101, 0b00100, 0b00100, 0b00100, 0b00000};
//const byte arrowDown[8] = {0b00000, 0b00100, 0b00100, 0b00100, 0b10101, 0b01110, 0b00100, 0b00000};

/* *******************************************************
   Serial Interface
 * **************************************************** */
// if you don't have enough HW Serial (i.e. on an UNO)
// you are forced to use SoftwareSerial or AltSoftSerial
#include <SoftwareSerial.h>
SoftwareSerial modbusSerial(2, 3);               // RXpin, TXpin

// On a Mega you can simply use
// a Reference to an existing HW Serial:
// HardwareSerial &modbusSerial = Serial3;

// If you don't need debugging output, you can even use
// a reference to Serial
// HardwareSerial &modbusSerial = Serial;

constexpr byte modbus_format = SERIAL_8N1;   // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;      // use slow speeds with SoftSerial
constexpr byte modbus_enable_pin = 11;       // The GPIO used to control the MAX485 TX pin. Set to 255 if you are not using RS485 or a selfsensing adapter

/* **************************************************** *
   Modbus
 * **************************************************** */

#include <SimpleModbusSlave.h>               // the adopted Modbus Slave library
#include <modbus_common.h>                   // some predefined Modbus registers
#include <modbus_slave_lcd.h>                // the Modbus Slave - LCD Adapter

constexpr byte modbus_slave_id = 2;          // Modbus address of this slave

// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  DI12,              // SI:
  DO13,              // SO: a digitalWrite to GPIO
  ADC0,              // SI: analogRead
  ADC1,              // SI: analogRead
  ADC2,              // SI: analogRead
  ADC3,              // SI: analogRead
  ADC4,              // SI: analogRead - be carefully, that is an I2C pin
  ADC5,              // SI: analogRead - be carefully, that is an I2C pin
  HOLDING_REGS_SIZE  // leave this one
  // total number of registers for function 3 and 16 share the same register array
  // i.e. the same address space
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

#include "LCD.h"
Indicator indicatorMasterOnline('|', 15, 1);        // character, col, row
Indicator indicatorMasterTimeout('x', 15, 1, 0);    // character, col, row, no timeout
//Indicator indicatorMuster(4, 15, 0, 0);           // no timeout --> 0 = permanent on

void onMasterActivity()
{
  indicatorMasterOnline.on();
  indicatorMasterTimeout.off();
}

/*
   this functions will be called,
   when a Modbus Master requests values from this Modbus Slave (FC3).
   If you don't need the debug messages, you can comment the function.

   This Example only shows a debug messages to indicate Master activity
*/
void modbusRegisterRequested(const uint16_t reg, const uint8_t count)
{
  Serial.print(F("requested:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
  onMasterActivity();
}

/*
   this functions will be called,
   when a Modbus Master modifes a value with FC6 or several values with FC16.
*/
void modbusRegisterModified(const uint16_t reg, const uint8_t count)
{
  Serial.print(F("modified:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
  onMasterActivity();
}



void updateRegisters()
{
  // write new values to GPIO
  digitalWrite(13, holdingRegs[DO13]);
  // update data to be read by the master
  holdingRegs[ADC0] = analogRead(A0);
  holdingRegs[ADC1] = analogRead(A1);
  holdingRegs[ADC2] = analogRead(A2);
  holdingRegs[ADC3] = analogRead(A3);
  holdingRegs[ADC4] = analogRead(A4);
  holdingRegs[ADC5] = analogRead(A5);
  holdingRegs[DI12] = digitalRead(12);
}




/*
   show some static information on LCD
*/
void showSplashscreen()
{
  lcd.clear();
  lcd.print(F(SPLASHSCREEN0));
  lcd.setCursor(0, 1);
  lcd.print(F(SPLASHSCREEN1));
  if (rows > 2)
  {
    lcd.setCursor(0, 2);
    lcd.print(F(SPLASHSCREEN2));
    lcd.setCursor(0, 3);
    lcd.print(F(SPLASHSCREEN3));
  }
}

/*
   show some static information on LCD
*/
void showConfig()
{
  lcd.clear();
  lcd.print(F("Addr:")); lcd.print(modbus_slave_id);

  lcd.setCursor(0, 1);
  lcd.print(F("Baud:")); lcd.print(modbus_baud);
  lcd.write(' ');
  switch (modbus_format)                 // from HardwareSerial.h - but only the most relevant
  {
    case 0x04 : lcd.print(F("7N1")); break;
    case 0x06 : lcd.print(F("8N1")); break;
    case 0x0C : lcd.print(F("7N2")); break;
    case 0x0E : lcd.print(F("8N2")); break;
    case 0x24 : lcd.print(F("7E1")); break;
    case 0x26 : lcd.print(F("8E1")); break;
    case 0x2C : lcd.print(F("7E2")); break;
    case 0x2E : lcd.print(F("8E2")); break;
    case 0x34 : lcd.print(F("7O1")); break;
    case 0x36 : lcd.print(F("8O1")); break;
    case 0x3C : lcd.print(F("7O2")); break;
    case 0x3E : lcd.print(F("8O2")); break; // ok now I understand the logik ;-)
    default:
      lcd.print(modbus_format);
  }
}


void handleButton()
{
  static int previousButton = -1;
  int currentButton = readLCDbuttons();
  if (currentButton != previousButton)
  {
    switch (currentButton)
    {
      case btnSELECT :
        showConfig();
    }
  }
}


/*
   time management for the LCD.
   currently it only deletes the splash screen after startup
*/
void lcdTimer()
{
  if (lcdWasUpdated == false && millis() > splashscreenTimeout * 1000UL) // delete splash screen
  {
    lcd.clear();
    lcdWasUpdated = true;
    indicatorMasterTimeout.display();          // assumption, nothing received yet
  }
}

void sanityTimer()
{
  // check if master is offline
  if (millis() - previousMasterMessage > 10000)
  {
    indicatorMasterTimeout.on();
  }
  //heartbeat
}


/* **************************************************** *
   Arduino default setup and loop
 * **************************************************** */
void setup()
{
  Serial.begin(115200);                 // only used for Debugging
  Serial.println(F("Modbus LCD Slave 0204"));

  modbusSerial.begin(modbus_baud);      // start the Serial for Modbus - Remember: Softserial doesn't support modbus_format
  //modbusSerial.begin(modbus_baud, modbus_format); // Hardware Serial supports modbus_format

  modbus_configure(&modbusSerial, modbus_baud, modbus_format, modbus_slave_id, modbus_enable_pin, HOLDING_REGS_SIZE, holdingRegs);

  lcd.begin();                         // initialize the display
  lcd.backlight();
  //lcd.createChar(5, arrowUp);
  //lcd.createChar(6, arrowDown);
  lcd.setCursor(0, 0);
  //lcd.print(F("LCD KeyPad Shield "));  // print a simple message
  if (splashscreenTimeout > 0) showSplashscreen();
  delay(1000);
  if (splashscreenTimeout > 0) showConfig();

  pinMode(13, OUTPUT);
  indicatorMasterTimeout.display();
}

void loop()
{
  modbus_update();   // call modbus in loop
  lcdTimer();        // call time management for the LCD
  updateRegisters(); // handle GPIOs (outputs like relais or LEDs and Inputs)
  handleButton();

  sanityTimer();

  indicatorMasterOnline.update();
  indicatorMasterTimeout.update();

  // do other things non blocking in loop
}
