/*******************************************************
  MAX7219 Modbus Slave (Receiver)

  PROTOTYPE ONLY

  hardware
  - a MAX7219 connected to the Arduino
  - a MAX485-TTL adapter 

  by noiasca
  2021-10-28 OK 

 *******************************************************/

#include <modbus_common.h>
#include <SimpleModbusSlave.h>

/*
   SimpleModbusSlaveV10 supports function 3, 6 & 16.

   This example code will receive the adc ch0 value from the arduino master.
   It will then use this value to adjust the brightness of the led on pin 9.
   The value received from the master will be stored in address 1 in its own
   address space namely holdingRegs[].

   In addition to this the slaves own adc ch0 value will be stored in
   address 0 in its own address space holdingRegs[] for the master to
   be read. The master will use this value to alter the brightness of its
   own led connected to pin 9.

   The modbus_update() method updates the holdingRegs register array and checks
   communication.

   Note:
   The Arduino serial ring buffer is 64 bytes or 32 registers.
   Most of the time you will connect the arduino to a master via serial
   using a MAX485 or similar.

   In a function 3 request the master will attempt to read from your
   slave and since 5 bytes is already used for ID, FUNCTION, NO OF BYTES
   and two BYTES CRC the master can only request 58 bytes or 29 registers.

   function 6: Presets a value into a single holding register (4X reference)

   In a function 16 request the master will attempt to write to your
   slave and since a 9 bytes is already used for ID, FUNCTION, ADDRESS,
   NO OF REGISTERS, NO OF BYTES and two BYTES CRC the master can only write
   54 bytes or 27 registers.

   Using a USB to Serial converter the maximum bytes you can send is
   limited to its internal buffer which differs between manufactures.
*/

#define  LED 13

// Using the enum instruction allows for an easy method for adding and
// removing registers. Doing it this way saves you #defining the size
// of your slaves register array each time you want to add more registers
// and at a glimpse informs you of your slaves register layout.

// holding registers (occupy memory)
enum
{
  // just add or remove registers and your good to go...
  // The first register starts at address 0 which is for Function 3 Register 40001
  ADC0,              // SI: analogRead of A0
  DI12,              // SI: 
  DO13,              // SO: a digitalWrite to GPIO
  HOLDING_REGS_SIZE  // leave this one
  // total number of registers for function 3 and 16 share the same register array
  // i.e. the same address space
};
unsigned int holdingRegs[HOLDING_REGS_SIZE]; // function 3 and 16 register array

// Display Hardware
constexpr byte ledCSPin =  8;
#include <NoiascaLedControlSpi.h>  // this is for Hardware SPI
LedControlSpi lc = LedControlSpi  (ledCSPin, 1);  // Hardware SPI


// Modbus (Interface) Hardware
/* Valid modbus byte formats are:
   SERIAL_8N2: 1 start bit, 8 data bits, 2 stop bits
   SERIAL_8E1: 1 start bit, 8 data bits, 1 Even parity bit, 1 stop bit
   SERIAL_8O1: 1 start bit, 8 data bits, 1 Odd parity bit, 1 stop bit

   You can obviously use SERIAL_8N1 but this does not adhere to the
   Modbus specifications. That said, I have tested the SERIAL_8N1 option
   on various commercial masters and slaves that were suppose to adhere
   to this specification and was always able to communicate... Go figure.

   These byte formats are already defined in the Arduino global name space.
*/
constexpr byte modbus_format = SERIAL_8N1;   // if you can use Hardware Serial, prefered value is SERIAL_8N2
constexpr uint32_t modbus_baud = 19200;
constexpr byte modbus_enable_pin = 5;
constexpr byte modbus_slave_id = 1;

// functions which will be called by the modbus slave library:

byte modbusWrite(const char actual)
{
  return lc.write(actual);
}

void modbusRegisterRequested(const uint16_t reg, const uint8_t count)
{
  Serial.print(F("requested:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

void modbusRegisterReceived(const uint16_t reg, const uint16_t value)
{
  Serial.print(F("received:")); Serial.print(reg); Serial.print(F(" value:")); Serial.println(value);
  switch (value)
  {
    case addrClear : lc.clearDisplay(0); break;
    case addrHome : lc.setCursor(0, 0); break;
    case addrSetBacklight : lc.setIntensity(0, value % 16);
  }
}

void modbusRegisterModified(const uint16_t reg, const uint8_t count)
{
  Serial.print(F("modified:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

void setup()
{
  Serial.begin(115200);
  Serial3.begin(modbus_baud, modbus_format);  // HW Serial, SoftSerial, AltSoftSerial,... need a begin method - call it
  /* parameters(HardwareSerial* SerialPort,
                long baudrate,
  	unsigned char byteFormat,
                unsigned char ID,
                unsigned char transmit enable pin,
                unsigned int holding registers size,
                unsigned int* holding register array)
  */
  modbus_configure(&Serial3, modbus_baud, modbus_format, modbus_slave_id, modbus_enable_pin, HOLDING_REGS_SIZE, holdingRegs);

  // modbus_update_comms(baud, byteFormat, id) is not needed but allows for easy update of the
  // port variables and slave id dynamically in any function.
  //modbus_update_comms(9600, SERIAL_8N2, 1);

  pinMode(LED, OUTPUT);

  // init the display
  setupMax7219();
}

void loop()
{
  // modbus_update() is the only method used in loop(). It returns the total error
  // count since the slave started. You don't have to use it but it's useful
  // for fault finding by the modbus master.
  modbus_update();

  holdingRegs[ADC0] = analogRead(A0); // update data to be read by the master to adjust the PWM
  //analogWrite(LED, holdingRegs[PWM_VAL] >> 2); // constrain adc value from the arduino master to 255
}
