/*
  Modbus Server LCD Receiver 
  -      -     -   -         
  
  MSLR
  
  The "LCD Receiver" receives Modbus Servers commands and translates them to LCD commands.
  
  Modbus Server --> Modbus Server LCD Receiver --> LCD Library
  
  Variables, definitions and callback functions specific to Modbus Server LCD Receiver
*/
#pragma once
#include "NoiascaModbusCommon.h"             // constants for example the addresses

/*
   This defines additional valid registers.
   This registers execute commands and are therefore not stored as "holding registers".
*/
bool mslrIsValidAddress(uint16_t addr)
{
  switch(addr)
  {
    case addrWrite :             
    case addrWriteOld :
    case addrBegin :
    case addrSetDelay :
    case addrCommand :
    case addrClear :
    case addrHome :
    case addrSetCursor :
    case addrCursor :
    case addrBlink :
    case addrSetBacklight :
    case addrStream :
    case addrStreamExpand : 
    case addrCreateChar :    
      return true;
  }
  return false;
}

bool mslrLcdWasUpdated = false;        // has the Server already received LCD changing commands
uint8_t mslrBuffer[8 + 1];             // we need a destination character and 8 bytes for special char
uint8_t mslrBufferIndex = 0;           // actual index of buffer

/*
   This functions will be called,
   when a Modbus Client sends a value with FC6/FC16 to a valid address (but not holding register)
   It acts as LCD Receiver converting Modbus Registers to LCD commands and writes
*/
void mslrRegisterSent(const uint16_t startAddress, const uint16_t value)
{
  // some debug information:
  //Serial.print(F("sent:")); Serial.print(startAddress); 
  //Serial.print(F(" value:")); Serial.print(value);
  //if(value > 0x20 && value <= 0xff) 
  //{
  //  Serial.print(' ');
  //  Serial.write(value);
  //}
  //Serial.println();
  
  byte hi = value >> 8;
  byte lo = value & 0xFF;
  switch (startAddress)
  {
    case addrStream:
      if (mslrLcdWasUpdated == false)      // before the first external character - delete content
      {
        lcd.clear();
        mslrLcdWasUpdated = true;
      }
      lcd.write(value >> 8);           // 2 bytes packet in one uint16_t
      lcd.write(value & 0xFF);
      break;
    case addrStreamExpand :
    case addrWrite :
      if (mslrLcdWasUpdated == false)      // before the first external character - delete content
      {
        lcd.clear();
        mslrLcdWasUpdated = true;
      }    
      lcd.write(value & 0xFF);
      break;
    case addrWriteOld :                // write any value (incl. special chars)
      lcd.writeOld(value & 0xFF);         
      break;
    case addrBegin : 
      lcd.begin(); 
      break;
    case addrClear :
      mslrLcdWasUpdated = true;
      lcd.clear();
      break;
    case addrHome :
      lcd.setCursor(0, 0);             // on most LCD faster than .home()
      break; 
    case addrCursor :
      if (lo == 0) 
        lcd.noCursor(); 
      else 
        lcd.cursor();
      break;
    case addrBlink :
      if (lo == 0) 
        lcd.noBlink(); 
      else 
        lcd.blink(); 
      break;
    case addrSetCursor :
      mslrLcdWasUpdated = true;
      lcd.setCursor(hi, lo);
      break;
    case addrSetBacklight : 
      if (lo == 0) 
        lcd.noBacklight(); 
      else 
        lcd.backlight(); 
      break;
    case addrCreateChar : 
      if (mslrBufferIndex < sizeof(mslrBuffer))
      {
        mslrBuffer[mslrBufferIndex] = lo;
        mslrBufferIndex++;
      }
    break;
    //default: 
      // valid holding registeres were already handled by the Modbus Library
  }
}

/*
   this functions can be called,
   when a Modbus Client modifes a value with FC6 or several values with FC16 to a holding register
   This example only shows a debug messages to indicate Client activity.   
*/
void mslrRegisterWritten(const uint16_t reg, const uint16_t value)
{
  Serial.print(F("written:")); Serial.print(reg); 
  Serial.print(F(" value:")); Serial.print(value); 
  Serial.print(F(" 0x")); Serial.print(value, HEX);
  Serial.println();
}

/*
   this functions can be called,
   when a Modbus Client requests values from this Modbus Server (FC3).
   If you don't need the debug messages don't set the callback.
   This example only shows a debug messages to indicate Client activity.
*/
void mslrRegisterRead(const uint16_t reg, const uint16_t count)
{
  Serial.print(F("read:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
}

/*
   this functions can be called,
   when a Modbus Client has finished the stream of data.
   It can be used to do an explicit call like "flush()".
   This example only shows a debug messages to indicate Client activity
*/
void mslrStreamEnd(const uint16_t reg, const uint16_t count)
{
  // Serial.print(F("stream end:")); Serial.print(reg); Serial.print(F(" count:")); Serial.println(count);
  (void)count;
  // we need it for the handling of special characters
  // the buffer contains 8 bitpatterns and in the last field the destination character
  if (reg == addrCreateChar)
  {
    lcd.createChar(mslrBuffer[8], mslrBuffer);
    mslrBufferIndex = 0;
  }
}
