/*
   Modbus Server (aka Modbus Slave) 
   
   credits:
   This library is inspired by "SimpleModbusSlaveV10" from Juan Bester.
   
   by noiasca
   2022-01-01   added support of FC2 Discrete Inputs, FC4 InputRegister, FC1/FC5/FC15 Coils
   2021-12-23

*/

#pragma once
#include <Arduino.h>

#define NOIASCA_MODBUS_SERVER_SIMPLE_VERSION "0.0.3"       //!> Version of this library

#define  MODBUS_DEBUG  0     //!> Turn on debug statements to the serial output

#if MODBUS_DEBUG
#define DEBUG(...) Serial.print(__VA_ARGS__)
#define DEBUGLN(...) Serial.println(__VA_ARGS__)
#else
#define DEBUG(...)
#define DEBUGLN(...) 
#endif  

/*
        Function                              Bit    registers
!   FC3 Read Multiple Holding register        16     start with 4 and span from 40001 to 49999 - supported 
!   FC6 Write Single Holding Register         16     start with 4 and span from 40001 to 49999 - supports additional registers
!   FC16 Write Multiple Holding Register      16     start with 4 and span from 40001 to 49999 - supports additional registers
!   FC2 Read discrete inputs                   1     start with 1 and span from 10001 to 19999 - supported
!   FC1 Read Coils                             1     start with 0 and span from 00001 to 09999 - supported
!   FC5 Write Single Coil                      1     start with 0 and span from 00001 to 09999 - supported
!   FC15 Write multiple Coils                  1     start with 0 and span from 00001 to 09999 - supported
!   FC4 Read input register                   16     start with 3 and span from 30001 to 39999 - supported
tbd:
    
    0x07 Read Exception Status
    0x08 Diagnostics
    FC23 Read/Write Multiple Registers        16                                                 not part of this lib
    FC22 Mask Write REgister                  16                                                 not part of this lib
    FC24 Read FIFO Queue                      16                                                 not part of this lib
*/

//#define BUFFER_SIZE 64              // replaced by an inclass static const...


/*
  \brief returns false for all given addresses
  
   this will be used as long as the user hasn't set it's own callback function
*/
bool cbInternalDummy(uint16_t address)
{
  (void)address;
  return false;
}

/**
  \brief Class for Modbus Server
   Modbus Server (aka Modbus Slave)
*/
class ModbusServer
{
  protected:
    static const byte BUFFER_SIZE = 64;
    uint8_t frame[BUFFER_SIZE];        // used to receive and transmit packages
    size_t holdingRegsSize = 1;        // size of the register array
    uint16_t* regs;                    // user array holding register
    uint8_t broadcastFlag;             // will be set if incoming message was a broadcast (server ID = 0)
    uint8_t slaveId;                   // id of server should be (1 .. 247)
    uint8_t function;                  // function code: usually the first byte in the Modbus message
    uint16_t errorCount = 0;           // legacy. very often CPT3
    uint16_t T1_5;                     // inter character time out
    uint16_t T3_5;                     // frame delay
    uint16_t noOfCoils = 0;            // no of coils for FC1/FC5/FC15
    uint16_t noOfDiscreteInputs = 0;   // no of discrete inputs for FC2
    uint16_t cpt1 = 0;                 // Bus Message Count
    uint16_t cpt2 = 0;                 // Bus Communication Error Count
    uint16_t cpt3 = 0;                 // Slave Exception Error Coun
    uint16_t cpt4 = 0;                 // Slave Message Count
    uint16_t cpt5 = 0;                 // Slave No Response Coun
    //uint16_t cpt6 = 0;                 // Slave NAK Count
    //uint16_t cpt7 = 0;                 // Slave Busy Count
    uint16_t cpt8 = 0;                 // Bus Character Overrun Cou
    Stream *ModbusPort;                // pointer to the used serial interface 
    
    // several callback pointers
    void (*cbPreTransmission)();       // gets called before writing a Modbus message
    void (*cbPostTransmission)();      // gets called after writing a Modbus message
    void (*cbOnClientActivitiy)();   
    bool (*cbIsValidAddress)(uint16_t); // validate additional addresses for FC3/FC6/FC16
    bool (*cbIsValidInputRegister)(uint16_t);    // validate input registers for FC4
    //using CallBack = void (*)(unsigned int startingAddress, unsigned int value);
    using CallBack = void (*)(uint16_t startingAddress, uint16_t value);
    CallBack cbRegisterWritten;        // FC6/FC16 write: Client has written to a server holding register
    CallBack cbRegisterSent;           // FC6/FC16 write: Client has sent value to an additional address
    CallBack cbRegisterRead;           // FC3 read:       Client has requested server holding registers
    uint16_t (*cbRegisterViewed)(uint16_t);// FC3 read:   Client has requested an additional address
    uint16_t (*cbInputRegisterViewed)(uint16_t);// FC4 read:   Client has requested an input register
    CallBack cbStreamEnd;              // FC16 write:     stream transmission has finished
    CallBack cbWriteCoil;              // FC5/FC15
    uint16_t (*cbReadCoil)(uint16_t);  // FC1 Read Coils on Modbus
    uint16_t (*cbReadDiscreteInput)(uint16_t);  // FC2 Read Discrete Inputs on Modbus

    void sendPacket(unsigned char bufferSize)
    {
      if(cbPreTransmission) cbPreTransmission();           // can be used to write HIGH to txEnablePin
      for (unsigned char i = 0; i < bufferSize; i++)
        (*ModbusPort).write(frame[i]);
      (*ModbusPort).flush();
      // allow a frame delay to indicate end of transmission
      delayMicroseconds(T3_5); 
      if(cbPostTransmission) cbPostTransmission();         // can be used to write LOW to txEnablePin
    }

/**
  \brief sends exception message

*/    
    void exceptionResponse(unsigned char exception)
    {
      if (exception <= 3) cpt3++;
      if (exception == 4) cpt4++; // processing error in application processing
      // each call to exceptionResponse() will increment the errorCount
      errorCount++; 
      if (!broadcastFlag)              // don't respond if its a broadcast message
      {
        frame[0] = slaveId;
        frame[1] = (function | 0x80);  // set MSB bit high, informs the master of an exception
        frame[2] = exception;
        unsigned int crc16 = calculateCRC(3);    // ID, function|0x80, exception code
        frame[3] = crc16 >> 8;
        frame[4] = crc16 & 0xFF;
        // exception response is always 5 bytes 
        // ID, function + 0x80, exception code, 2 bytes crc
        sendPacket(5); 
      }
      DEBUG (F("E129 exception "));
      DEBUG (exception);
      switch(exception)            // only for the 4 major exceptions we write clear text
      {
        case 1 :
        DEBUG(F(" ILLEGAL FUNCTION"));
        break;
        case 2 :
        DEBUG(F(" ILLEGAL DATA ADDRESS"));
        break;
        case 3 :
        DEBUG(F(" ILLEGAL DATA VALUE"));
        break;
        case 4 :
        DEBUG(F(" SLAVE DEVICE FAILURE"));
        break; 
      }
      DEBUG (F(" in FC"));
      DEBUGLN(function);
    }
    
/**
  \brief calculates CRC of bufferSize
  
  @param  bufferSize the used number of bytes in the frame buffer
  @returns caclulated CRC
*/      
    unsigned int calculateCRC(unsigned char bufferSize) 
    {
      unsigned int temp, temp2, flag;
      temp = 0xFFFF;
      for (unsigned char i = 0; i < bufferSize; i++)
      {
        temp = temp ^ frame[i];
        for (unsigned char j = 1; j <= 8; j++)
        {
          flag = temp & 0x0001;
          temp >>= 1;
          if (flag)
            temp ^= 0xA001;
        }
      }
      // Reverse byte order. 
      temp2 = temp >> 8;
      temp = (temp << 8) | temp2;
      temp &= 0xFFFF; 
      // the returned value is already swapped
      // crcLo byte is first & crcHi byte is last
      return temp; 
    }    
    
  public:
/**
  \brief This is the constructor of the ModbusServer Simple class
  
  @param serverId the server ID (slave ID) of this device 
  @param serial a reference to the serial interface. Can be an instance of HW Serial, SoftSerial or AltSerial.
*/ 
  
    ModbusServer (uint8_t serverId, Stream &serial) : 
      slaveId {serverId},
      ModbusPort {&serial}
    {
      cbPostTransmission = 0;
      cbPreTransmission = 0;
      cbOnClientActivitiy = 0;
      cbRegisterWritten = 0;
      cbRegisterRead = 0;
      cbRegisterViewed = 0;
      cbIsValidAddress = cbInternalDummy;
      cbInputRegisterViewed = 0;
      cbIsValidInputRegister = cbInternalDummy;
      cbRegisterSent = 0;
      cbStreamEnd = 0;
      cbWriteCoil = 0;
      cbReadCoil = 0;
      cbReadDiscreteInput = 0;      
    }
    
    /**
      \brief action before the first byte will be transmitted.
      
      This function gets called just before a Modbus message is responded over serial.
      Typical usage of this callback is to enable an RS485 transceiver's
      Driver Enable pin, and optionally disable its Receiver Enable pin.
    */
    void setPreTransmission(void (*cbPreTransmission)())
    {
      (*this).cbPreTransmission = cbPreTransmission;
    }

    /**
      \brief action after the last byte was transmitted.
      
      This function gets called after a Modbus message has finished sending
      (i.e. after all data has been physically transmitted onto the serial bus).

      Typical usage of this callback is to enable an RS485 transceiver's
      Receiver Enable pin, and disable its Driver Enable pin.
    */
    void setPostTransmission(void (*cbPostTransmission)())
    {
      (*this).cbPostTransmission = cbPostTransmission;
    }
   
    /**
      \brief validate received additional addresses.
      FC3/FC6/FC16.
      
      This function validates should validate if the received address is correct.
      
      @see setRegisterSent
      @see setRegisterWritten
      @see setRegisterViewed      
    */
    void setIsValidAddress(bool (*cbIsValidAddress)(uint16_t))
    {
      (*this).cbIsValidAddress = cbIsValidAddress;
    }
    
    /**
      \brief validate input register.
      FC4 Read input register
      
      Set callback to to validate if the address should be handled as valid input register.
      @see setInputRegisterViewed      
      
    */
    void setIsValidInputRegister(bool (*cbIsValidInputRegister)(uint16_t))
    {
      (*this).cbIsValidInputRegister = cbIsValidInputRegister;
    }

    /**
      \brief Client has sent an additional register 
      FC6/FC16 Write Holding Registers if Holding Register is an additional address.
      
      Set callback to take over received data to user sketch
      
      @see setIsValidAddress
      @see setRegisterWritten
      @see setRegisterViewed
    */
    void setRegisterSent(const CallBack cbRegisterSent)
    {
      (*this).cbRegisterSent = cbRegisterSent;
    }
    
    /**
      \brief Client has written to a holding register.
      FC6/FC16 Write Holding Register.
      
      Set callback to acknowledge the user sketch about client write activity
      
      @see setRegisterRead inverse function for FC3
    */
    void setRegisterWritten(const CallBack cbRegisterWritten)
    {
      (*this).cbRegisterWritten = cbRegisterWritten;
    }  
    
    /**
      \brief Client has read a holding register.
      FC3 Read Multiple Holding register.
      
      Set callback to acknowledge the user sketch about client read activity.
      @see setRegisterWritten inverse function for FC6/FC16
    */
    void setRegisterRead(const CallBack cbRegisterRead)
    {
      (*this).cbRegisterRead = cbRegisterRead;
    }

    /**
      \brief Client has viewed an additional address
      FC3 Read Multiple Holding Register. 
      
      Set callback to read from an additional address (outside of Holding Register Array).
      
      @see setIsValidAddress
      @see setRegisterSent inverse Function for FC6/16
    */
    void setRegisterViewed(uint16_t (*cbRegisterViewed)(uint16_t))
    {
      (*this).cbRegisterViewed = cbRegisterViewed;
    }
    
    /**
      \brief Client has viewed an input register.
      FC4 Read Input Register. 
      
      Set callback to receive value from input register.
      
      @see setIsValidInputRegister
    */
    void setInputRegisterViewed(uint16_t (*cbInputRegisterViewed)(uint16_t))
    {
      (*this).cbInputRegisterViewed = cbInputRegisterViewed;
    }
    
    /**
      \brief sets the valid number of coils
      FC1/FC5/FC15
      
      FC1 Read Coils.
      FC5 Write Single Coil.
      FC15 Write Multiple Coils.
      @return 0 on success
      @see setWriteCoil
      @see setReadCoil  
    */   
    int setNoOfCoils(uint16_t newValue)
    {
     if (newValue <= 0x7D0) // Coils 1 to 2000. 0 means "no coils"
     {
       noOfCoils = newValue;
       return 0;
     }
     else return -1;        // error
    }  

    /**
      \brief returns one readed Coil from user sketch
      FC1 Read Coils.
      
      Set callback to retrieves coil status from user sketch.
      @see setNoOfCoils
      @see setWriteCoil inverse function.
    */    
    void setReadCoil(uint16_t (*cbReadCoil)(uint16_t))
    {
      (*this).cbReadCoil = cbReadCoil;
    }    
    
    /**
      \brief Writes value to coil.
      FC5 Write Single Coil. 
      FC15 Write Multiple Coils.
      
      Set callback to write coil commands. User must implement the On or Off of the coil (output).
      @see setReadCoil inverse function.
    */
    void setWriteCoil(const CallBack cbWriteCoil)
    {
      (*this).cbWriteCoil = cbWriteCoil;
    }
    
    /**
      \brief returns one discrete input from user sketch.
      FC2 Read Discrete Input.
      
      Set callback to retrieves discrete input status from user sketch.
      @see setNoOfDiscreteInputs
    */
    void setReadDiscreteInput(uint16_t (*cbReadDiscreteInput)(uint16_t))
    {
      (*this).cbReadDiscreteInput = cbReadDiscreteInput;
    }
    
    /**
      \brief transmission of several registers has ended.
      FC16 Write Multiple Hodling Registers.
      
      Set callback to function when several values were send using FC16 / end of transmission of several values.
      
    */   
    void setStreamEnd(const CallBack cbStreamEnd)
    {
      (*this).cbStreamEnd = cbStreamEnd;
    }

    /**
      \brief Client activity.
      
      Set callback to a function which gets called if client has send a to this server.
    */   
    void setOnClientActivity(void (*cbOnClientActivitiy)())
    {
      (*this).cbOnClientActivitiy = cbOnClientActivitiy;
    }
  
    /**
      \brief sets the valid number of discrete inputs
      FC2 Read Discrete Input
      
      Define the number of discrete Inputs.
      
      @return 0 on success
      @see setReadDiscreteInput
    */  
    int setNoOfDiscreteInputs(uint16_t newValue)
    {
     if (newValue <= 0x7D0) // Discrete Inputs 1 to 2000. 0 means "no discrete inputs"
     {
       noOfDiscreteInputs = newValue;
       return 0;
     }
     else return -1;        // error
    }
    
    /**
       \brief configures the Modbus instance
       
       Init function. Call this function in setup() 
       former modbus_configure
       
       @param SerialPort a pointer to your serial interface.
       @param baud the baud rate of your serial interface.
       @param byteFormat e.g. 8N1
       @param serverId the server ID (slave ID) of your device.
       @param holdingRegsSize the total number of all holding registers.
       @param regs a pointer to your array of holding registers.
    */
    void begin (Stream *SerialPort,
                long baud,
                uint8_t byteFormat,
                uint8_t serverId, 
                uint16_t holdingRegsSize,
                uint16_t* regs)
    {
      ModbusPort = SerialPort;
      modbus_update_comms(baud, byteFormat);
      setServerId(serverId);
      (*this).holdingRegsSize = holdingRegsSize; 
      (*this).regs = regs;
      errorCount = 0; // initialize errorCount
    } 

    /**
      \brief modify Modbus parameters during runtime
      
      Can be used to modify Modbus parameters. Will adopt frame rates based on baud.
      Don't forget to set baud (and byte format) of your interface.
      
      @param baud   speed of serial interface.
      @param byteFormat e.g. 8N1      
    */       
    void modbus_update_comms(long baud, unsigned char byteFormat)
    {
      //(*ModbusPort).begin(baud, byteFormat);  // must be called in Usersketch because only Hardware Serial support byteFormat
      (void)byteFormat;
      
      // Modbus states that a baud rate higher than 19200 must use a fixed 750 us 
      // for inter character time out and 1.75 ms for a frame delay for baud rates
      // below 19200 the timing is more critical and has to be calculated.
      // E.g. 9600 baud in a 10 bit packet is 960 characters per second
      // In milliseconds this will be 960characters per 1000ms. So for 1 character
      // 1000ms/960characters is 1.04167ms per character and finally modbus states
      // an inter-character must be 1.5T or 1.5 times longer than a character. Thus
      // 1.5T = 1.04167ms * 1.5 = 1.5625ms. A frame delay is 3.5T.
      
      if (baud > 19200)
      {
        T1_5 = 750; 
        T3_5 = 1750; 
      }
      else 
      {
        T1_5 = 15000000 / baud; // 1T * 1.5 = T1.5
        T3_5 = 35000000 / baud; // 1T * 3.5 = T3.5
      }
    }  
    
    /**
      \brief modifies the Modbus Server ID (slave ID)
      
      Assigns a new Modbus server ID to this device. Currently the new server ID is not check, but don't use other values than 1..247.
      @param serverId the server ID (slave ID) from 1 to 247. 
    */
    void setServerId(uint8_t serverId)
    {
      slaveId = serverId;
    }
    
    /**
      \brief Main "run" function to be called in loop()
      
      former modbus_unsigned intupdate()
      @return errorCount
    */   
    uint16_t update()
    {
      if ((*ModbusPort).available())
      {
        uint8_t buffer = 0;
        uint8_t overflow = 0;
      
        while ((*ModbusPort).available())
        {
          // If more bytes are received than the BUFFER_SIZE the overflow flag will be set and the 
          // serial buffer will be red until all the data is cleared from the receive buffer.
          if (overflow) 
            (*ModbusPort).read();
          else
          {
            if (buffer == BUFFER_SIZE)
              overflow = 1;
            frame[buffer] = (*ModbusPort).read();
            buffer++;
          }
          delayMicroseconds(T1_5); // inter character time out
        }
      
        // If an overflow occurred increment the errorCount
        // variable and return to the main sketch without 
        // responding to the request i.e. force a timeout
        if (overflow)
          return errorCount++;
      
        // The minimum request packet is 8 bytes for function 3 & 16
        if (buffer > 7) 
        {
          uint8_t id = frame[0];
          broadcastFlag = 0;
    
          if (id == 0)
            broadcastFlag = 1;
        
          if (id == slaveId || broadcastFlag) // if the received ID matches the slaveId or broadcasting id (0), continue
          {
            uint16_t crc = ((frame[buffer - 2] << 8) | frame[buffer - 1]); // combine the crc Low & High bytes
            if (calculateCRC(buffer - 2) == crc) // if the calculated crc matches the received crc continue
            {
              cpt1++;
              function = frame[1];
              uint16_t startingAddress = ((frame[2] << 8) | frame[3]); // combine the starting address bytes
              uint16_t no_of_registers = ((frame[4] << 8) | frame[5]); // combine the number of register bytes  
              uint16_t maxData = startingAddress + no_of_registers;    // calculated last address to read
              uint8_t address; // was unsigned char - not so much an address than a different counter within the received bytes
              uint16_t crc16;
              if (id == 0 || id == slaveId) cpt4++;
              if (id == 0) cpt5++;
            
              // broadcasting makes no sense for function 1 
              if (!broadcastFlag && (function == 1))           // FC1 Read Coils
              {
                DEBUGLN(F("D404 FC1 received"));
                //Serial.print(F("startingAddress=   ")); Serial.println(startingAddress);
                //Serial.print(F("noOfCoils=")); Serial.println(noOfCoils);
                if (1 <= no_of_registers &&  no_of_registers <= 0x07D0)        // check exception 3 ILLEGAL DATA VALUE  
                {
                  if (startingAddress < noOfCoils && maxData <= noOfCoils)     // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if (cbReadCoil)                  
                    {
                      uint8_t noOfBytes = no_of_registers / 8 + 1;  // bitwise: d.h. angeforderte / 8?
                      // ID, function, noOfBytes, (number of registers / 8) + 1, crcLo, crcHi
                      uint8_t responseFrameSize = 5 + noOfBytes; 
                      frame[0] = slaveId;
                      frame[1] = function;
                      frame[2] = noOfBytes;
                      address = 3;               // PDU starts at the 4th byte
                      uint8_t count = 0;         // bit position
                      frame[address] = 0;
                      for (uint16_t index = startingAddress; index < maxData; index++)
                      {
                        if (count >= 7)          // next frame
                        {
                          address++;
                          frame[address] = 0;
                          count = 0;
                        }
                        if (cbReadCoil(index))          // existence of cb was checked already
                          frame[address] |= 1 << count;
                        count++;
                      }
                      crc16 = calculateCRC(responseFrameSize - 2);
                      frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                      frame[responseFrameSize - 1] = crc16 & 0xFF;
                      sendPacket(responseFrameSize);
                      
                      if (cbOnClientActivitiy) cbOnClientActivitiy();
                    }
                    else
                    {
                      exceptionResponse(4);          // exception 4 can't read coil state
                    }
                  }
                  else
                  {
                    exceptionResponse(2);          // exception 2 ILLEGAL DATA ADDRESS                    
                  }
                }
                else
                {
                  exceptionResponse(3);        // exception 3 ILLEGAL DATA VALUE
                }
              }
              
              // broadcasting is not supported for function 2 
              else if (!broadcastFlag && (function == 2))           // FC2 Read Discrete Inputs
              {
                DEBUGLN(F("D465 FC2 received"));
                //Serial.print(F("startingAddress=   ")); Serial.println(startingAddress);
                //Serial.print(F("noOfDiscreteInputs=")); Serial.println(noOfDiscreteInputs);
                if (1 <= no_of_registers &&  no_of_registers <= 0x07D0)      // check exception 3 ILLEGAL DATA VALUE  
                {
                  if (startingAddress < noOfDiscreteInputs && maxData <= noOfDiscreteInputs)     // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if(cbReadDiscreteInput)
                    {
                      uint8_t noOfBytes = no_of_registers / 8 + 1;  // bitwise: d.h. angeforderte / 8?
                      // ID, function, noOfBytes, (number of registers / 8) + 1, crcLo, crcHi
                      uint8_t responseFrameSize = 5 + noOfBytes; 
                      frame[0] = slaveId;
                      frame[1] = function;
                      frame[2] = noOfBytes;
                      address = 3; // PDU starts at the 4th byte         
                      uint8_t count = 0;  // bit position
                      frame[address] = 0;
                      for (uint16_t index = startingAddress; index < maxData; index++)
                      {
                        if (count >= 7)  // next frame
                        {
                          address++;
                          frame[address] = 0;
                          count = 0;
                        }
                        if (cbReadDiscreteInput(index))          // existence of cb was checked already
                          frame[address] |= 1 << count;
                        count++;
                      }

                      crc16 = calculateCRC(responseFrameSize - 2);
                      frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                      frame[responseFrameSize - 1] = crc16 & 0xFF;
                      sendPacket(responseFrameSize);
                      
                      if (cbOnClientActivitiy) cbOnClientActivitiy();
                    }
                    else
                    {
                      exceptionResponse(4);      // exception 4
                    }
                  }
                  else
                  {
                    exceptionResponse(2);        // exception 2 ILLEGAL DATA ADDRESS                    
                  }
                }
                else
                {
                  exceptionResponse(3);          // exception 3 ILLEGAL DATA VALUE
                }
              }

              // broadcasting is not supported for function 3 
              else if (!broadcastFlag && (function == 3))           // FC3 Read Multiple Holding Registers
              {
                DEBUGLN(F("D522 FC3 received"));
                //Serial.print(F("startingAddress=")); Serial.println(startingAddress);
                //Serial.print(F("holdingRegsSize=")); Serial.println(holdingRegsSize);
                if (startingAddress < holdingRegsSize || cbIsValidAddress(startingAddress))         // check exception 2 ILLEGAL DATA ADDRESS
                {
                  if (maxData <= holdingRegsSize || cbIsValidAddress(startingAddress))              // check exception 3 ILLEGAL DATA VALUE
                  {
                    uint8_t noOfBytes = no_of_registers * 2; 
                    // ID, function, noOfBytes, (dataLo + dataHi)*number of registers, crcLo, crcHi
                    uint8_t responseFrameSize = 5 + noOfBytes; 
                    frame[0] = slaveId;   
                    frame[1] = function;
                    frame[2] = noOfBytes;
                    address = 3; // PDU starts at the 4th byte
                    uint16_t temp = 0;
                    if (maxData <= holdingRegsSize)                                    // default: holding register
                    {
                      for (uint16_t index = startingAddress; index < maxData; index++)
                      {
                        temp = regs[index];
                        frame[address] = temp >> 8; // split the register into 2 bytes
                        address++;
                        frame[address] = temp & 0xFF;
                        address++;
                      }
                    }
                    else                                                                 // additional registers
                    {
                      for (uint16_t index = startingAddress; index < (startingAddress + no_of_registers); index++)
                      {                     
                        // MISSING tbd: check index against cbIsValidAddress
                        if (cbRegisterViewed) temp = cbRegisterViewed(index);   
                        frame[address] = temp >> 8; // split the register into 2 bytes
                        address++;
                        frame[address] = temp & 0xFF;
                        address++;
                      }
                    }                  

                    crc16 = calculateCRC(responseFrameSize - 2);
                    frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                    frame[responseFrameSize - 1] = crc16 & 0xFF;
                    sendPacket(responseFrameSize);
                    
                    if (maxData <= holdingRegsSize && cbRegisterRead) cbRegisterRead(startingAddress, no_of_registers);
                    if (cbOnClientActivitiy) cbOnClientActivitiy();
                  }
                  else
                  {                   
                    exceptionResponse(3);        // exception 3 ILLEGAL DATA VALUE
                  }
                }
                else
                {
                  exceptionResponse(2);          // exception 2 ILLEGAL DATA ADDRESS
                }
              }

              // broadcasting is not supported for function 4 
              else if (!broadcastFlag && (function == 4))           // FC4 Read Input Registers
              {
                DEBUGLN(F("D735 FC4 received"));
                //Serial.print(F("startingAddress=")); Serial.println(startingAddress);
                //Serial.print(F("holdingRegsSize=")); Serial.println(holdingRegsSize);
                uint16_t lastAddress = startingAddress + no_of_registers - 1;
                if (0x0001 <= no_of_registers && no_of_registers <= 0x007D)    // check exception 3 ILLEGAL DATA VALUE
                {
                  if (cbIsValidInputRegister(startingAddress) && cbIsValidInputRegister(lastAddress))        // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if (cbInputRegisterViewed)  
                    {  
                      uint8_t noOfBytes = no_of_registers * 2; 
                      // ID, function, noOfBytes, (dataLo + dataHi)*number of registers, crcLo, crcHi
                      uint8_t responseFrameSize = 5 + noOfBytes; 
                      frame[0] = slaveId;   
                      frame[1] = function;
                      frame[2] = noOfBytes;
                      address = 3; // PDU starts at the 4th byte
                      uint16_t temp = 0;
                      for (uint16_t index = startingAddress; index < (startingAddress + no_of_registers); index++)
                      {                     
                        temp = cbInputRegisterViewed(index);   
                        frame[address] = temp >> 8; // split the register into 2 bytes
                        address++;
                        frame[address] = temp & 0xFF;
                        address++;
                      }

                      crc16 = calculateCRC(responseFrameSize - 2);
                      frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                      frame[responseFrameSize - 1] = crc16 & 0xFF;
                      sendPacket(responseFrameSize);
                      
                      if (cbOnClientActivitiy) cbOnClientActivitiy();
                    }
                    else
                    {                 
                    exceptionResponse(4);
                    }                      
                  }
                  else
                  {                
                    exceptionResponse(2);          // exception 2 ILLEGAL DATA ADDRESS
                  }
                }
                else
                {
                  exceptionResponse(3);        // exception 3 ILLEGAL DATA VALUE
                }
              }

              // MISSING: tbd: allow broadcasting for function 5?
              else if (!broadcastFlag && (function == 5))           // FC5 Write Single Coil
              {
                DEBUGLN(F("D585 FC5 received"));
                //Serial.print(F("startingAddress=   ")); Serial.println(startingAddress);
                //Serial.print(F("noOfCoils=")); Serial.println(noOfCoils);
                
                uint16_t outputValue = frame[4] << 8 | frame[5];
                //  if (1 <= no_of_registers &&  no_of_registers <= 0x07D0)
                if (outputValue == 0x0000 || outputValue == 0xFF00)        // check exception 3 ILLEGAL DATA VALUE  
                {
                  if (startingAddress < noOfCoils)         // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if (cbWriteCoil)                       // cb is needed to be able to write to the coil
                    {
                      // ID, function, addrhi, addrlo, statHi, statLo, crcLo, crcHi
                      // response is simple echo of incomming frame[]
                      uint8_t responseFrameSize = 8; 
                      cbWriteCoil(startingAddress, outputValue);
                        
                      crc16 = calculateCRC(responseFrameSize - 2);
                      frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                      frame[responseFrameSize - 1] = crc16 & 0xFF;
                      sendPacket(responseFrameSize);
                      
                      if (cbOnClientActivitiy) cbOnClientActivitiy();
                    }
                    else
                    {
                      exceptionResponse(4);          // exception 4 can't write coil state
                    }
                  }
                  else
                  {
                    exceptionResponse(2);          // exception 2 ILLEGAL DATA ADDRESS                    
                  }
                }
                else
                {
                  exceptionResponse(3);        // exception 3 ILLEGAL DATA VALUE
                }
              }              
              
              else if (function == 6)            // FC6 Write Single Holding Register
              {
                DEBUGLN(F("D629 FC6 received"));
                if (startingAddress < holdingRegsSize || cbIsValidAddress(startingAddress)) // check exception 2 ILLEGAL DATA ADDRESS
                {
                  if (startingAddress < holdingRegsSize)
                  {
                    regs[startingAddress] = ((frame[4] << 8) | frame[5]);      // the 4th and 5th elements in frame is the 16 bit data value
                    if (cbRegisterWritten) cbRegisterWritten(startingAddress, 1);
                    if (cbOnClientActivitiy) cbOnClientActivitiy();
                  }
                  else if (cbIsValidAddress(startingAddress))
                  {
                    if (cbRegisterSent) cbRegisterSent(startingAddress, ((uint16_t)frame[4] << 8) | frame[5]);
                    if (cbOnClientActivitiy) cbOnClientActivitiy();
                  }
                  // only the first 6 bytes are used for CRC calculation
                  crc16 = calculateCRC(6); 
                  frame[6] = crc16 >> 8;         // split crc into 2 bytes
                  frame[7] = crc16 & 0xFF;
                  // a function 16 response is an echo of the first 6 bytes from 
                  // the request + 2 crc bytes
                  if (!broadcastFlag) // don't respond if it's a broadcast message
                    sendPacket(8);
                }
                else
                  exceptionResponse(2); // exception 2 ILLEGAL DATA ADDRESS
              }

              // MISSING: tbd: allow broadcasting for function 15?
              else if (!broadcastFlag && (function == 15))           // FC15 Write Multiple Coils
              {
                DEBUGLN(F("D660 FC15 received"));
                //Serial.print(F("startingAddress=   ")); Serial.println(startingAddress);
                //Serial.print(F("noOfCoils=")); Serial.println(noOfCoils);
                // uint16_t quantity = frame[4] << 8 | frame[5];
                //  Missing: check for byte Count = N*
                if (1 <= no_of_registers &&  no_of_registers <= 0x07D0)        // check exception 3 ILLEGAL DATA VALUE  
                {
                  if (startingAddress < noOfCoils && maxData <= noOfCoils)         // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if (cbWriteCoil)                       // cb is needed to be able to write to the coil
                    {
                      // request:
                      // ID, FC, addrhi, addrlo, quantityHi, quantityLo, Bytes, N * Output Value, crcLo, crcHi, 
                      // response:
                      // ID, FC, addrhi, addrlo, quantityHi, quantityLo, crcLo, crcHi, 
                      uint8_t responseFrameSize = 8; 
                      //frame[0] = slaveId;    // not needed as broadcast should not respond anyway
                      byte actualByte = 7;
                      uint8_t count = 0;     // bit position
                      for (uint16_t index = startingAddress; index < maxData; index++)
                      {
                        if (count >= 7)      // next frame
                        {
                          actualByte++;
                          count = 0;
                        }
                        if (frame[actualByte] & (1 << count)) 
                          cbWriteCoil(index, 1); // alternative: send 0xFF00 like FC5
                        else
                          cbWriteCoil(index, 0);
                        count++;
                      }
                      
                      crc16 = calculateCRC(responseFrameSize - 2);
                      frame[responseFrameSize - 2] = crc16 >> 8; // split crc into 2 bytes
                      frame[responseFrameSize - 1] = crc16 & 0xFF;
                      sendPacket(responseFrameSize);
                      
                      if (cbOnClientActivitiy) cbOnClientActivitiy();
                    }
                    else
                    {
                      exceptionResponse(4);          // exception 4 can't write coil state
                    }
                  }
                  else
                  {
                    exceptionResponse(2);          // exception 2 ILLEGAL DATA ADDRESS                    
                  }
                }
                else
                {
                  exceptionResponse(3);        // exception 3 ILLEGAL DATA VALUE
                }
              }                
              
              else if (function == 16)     // FC16 Write Multiple Holding Registers
              {
                // Check if the received number of bytes matches the calculated bytes 
                // minus the request bytes.
                // id + function + (2 * address bytes) + (2 * no of register bytes) + 
                // byte count + (2 * CRC bytes) = 9 bytes
                DEBUGLN(F("D725 FC16 received"));
                if (frame[6] == (buffer - 9)) 
                {
                  if (startingAddress < holdingRegsSize || cbIsValidAddress(startingAddress))      // check exception 2 ILLEGAL DATA ADDRESS
                  {
                    if (maxData <= holdingRegsSize || cbIsValidAddress(startingAddress)) // check exception 3 ILLEGAL DATA VALUE
                    {
                      address = 7;     // start at the 8th byte in the frame
                      if (startingAddress < holdingRegsSize)                             // default receive data - write received data to regs
                      {  
                        int modified = 0;
                        for (uint16_t index = startingAddress; index < maxData; index++)
                        {
                          regs[index] = ((frame[address] << 8) | frame[address + 1]);
                          address += 2;
                          modified++;
                        }
                        if (cbRegisterWritten) cbRegisterWritten(startingAddress, modified);       // to be checked
                        if (cbOnClientActivitiy) cbOnClientActivitiy();
                      }
                      else                                                               // use extended address room - callback
                      {
                        DEBUG(F("D747 current buffer")); DEBUGLN(buffer); 
                        for (int index = 7; index < buffer - 2; index++)
                        {              
                          DEBUG(" 0x");
                          DEBUG(frame[index], HEX);
                          DEBUG('=');
                          #if MODBUS_DEBUG
                          Serial.write(frame[index]);
                          #endif
                          if(cbRegisterSent) cbRegisterSent(startingAddress, frame[index] << 8 | frame[index + 1]);                            
                          index++;
                        }                   
                        DEBUGLN();
                        if (cbStreamEnd) cbStreamEnd(startingAddress, buffer - 7);     // evtl noch ein "StreamEnded"     
                        if (cbOnClientActivitiy) cbOnClientActivitiy();
                      }
    
                      // only the first 6 bytes are used for CRC calculation
                      crc16 = calculateCRC(6); 
                      frame[6] = crc16 >> 8;// split crc into 2 bytes
                      frame[7] = crc16 & 0xFF;
                    
                      // a function 16 response is an echo of the first 6 bytes from 
                      // the request + 2 crc bytes
                      if (!broadcastFlag)  // don't respond if it's a broadcast message
                        sendPacket(8); 
                    }
                    else
                    {                  
                      exceptionResponse(3);// exception 3 ILLEGAL DATA VALUE
                    }
                  }
                  else
                  {  
                    exceptionResponse(2);  // exception 2 ILLEGAL DATA ADDRESS
                  }
                }
                else 
                  errorCount++;            // corrupted packet
              }          
              else
                exceptionResponse(1);      // exception 1 ILLEGAL FUNCTION
            }
            else // checksum failed
            {
              cpt2++;
              errorCount++;
            }
          } // incorrect id
        }
        else if (buffer > 0 && buffer < 8)
          errorCount++; // corrupted packet
      }
      return errorCount;
    }  
};
