/*
  Noiasca Liquid Crystal 2.x - Modbus
  
  A LCD class for a Modbus Master.
  It doesn't drive a LCD but writes commmands via Modbus Serial Communication
  
  Variant: 
  - full LCD API interface - doesn't need another LCD Library
  - uses BufferPrint
  
  Hardware
  - Modbus Interface 
  
  Description
  - This LCD class provides the common LCD API 1.0 to a Modbus Master for transmitting LCD Data to a slave
  - This LCD class inherits from a special BufferPrint class.
    The single characters of a print will be buffered and only transmitted by one final transaction (FC16 Write Multiple Holding Registers).
    This fastens the transmission time over Modbus (19200) by factor 6.
  - write (and writeOld) are existing as single value transmission. 
    Only one byte (in a 16bit field) is transmitted (FC6 Write Single Holding Register) --> Logic addrStreamExpand
  - Dataflow: lcd_Modbus_BufferPrint (using BufferPrint) --> ModbusMaster --> Serial/RS485
    
  Open topics
  - some legacy LCD methods are not supported currently
  - special character support
  
  copyright 2021 noiasca noiasca@yahoo.com
  
  2021-11-03 ok
*/

#pragma once
#include <Arduino.h>
#include <ModbusMaster.h>              // "Version 2.0.0" - must also be included in user sketch
#include <NoiascaBufferPrint.h>        // improved print class for buffers print and collects several writes

//#define DEBUG              // activates debug messages

#ifdef DEBUG
  #define DEBUG_PRINTLN(x)  Serial.println(x)
#else
  #define DEBUG_PRINTLN(x)
#endif

// this Library needs following constants in the Usersketch 
// or included via a separate .h file

//constexpr uint16_t addrWrite =          ;             
//constexpr uint16_t addrWriteOld =       ; 
//constexpr uint16_t addrBegin =          ;
//constexpr uint16_t addrSetDelay =       ;
//constexpr uint16_t addrCommand =        ;
//constexpr uint16_t addrClear =          ;
//constexpr uint16_t addrHome =           ;
//constexpr uint16_t addrSetCursor =      ;
//constexpr uint16_t addrCursor =         ;
//constexpr uint16_t addrBlink =          ;
//constexpr uint16_t addrSetBacklight =   ;
//constexpr uint16_t addrStream =         ;  // FC6/FC16 - but mainly FC16 - 2 bytes in one 16bit - prints currently all characters breaks with odd characters
//constexpr uint16_t addrStreamExpand =   ;  // FC6/FC16 prints only low nibble from value ein Byte/Zeichen pro 16bit register - - Umstellen auf 1100?
//constexpr uint16_t addrCreateChar =     ;  

// backward compatibilty 
// for begin method:
// #define LCD_5x10DOTS 0x04                        // F  5x10
// #define LCD_5x8DOTS 0x00                         // F  5x7
//typedef enum {POSITIVE, NEGATIVE, FAULTY} t_backlightPol; // compatible to the very common New Liquid Crystal

/*
  
   A LCD like class - to be used as Modbus Client (Modbus Master) to SEND LCD data to a Modbus Server (Modbus Slave)
   
*/
class LiquidCrystal_Modbus : public NoiascaBufferPrint {

  protected:
    // Pin control constants
    //static const byte rsIR = 0;                            // register select: Instruction Register (commando)
    //static const byte rsDR = 1;                            // register select: Data Register
    static const byte rwWrite = 0;                         // read/write WRITE
    static const byte rwRead = 1;                          // read/write READ
    byte wait = 5;                                         // use 5 for 9600 Baud - set some delay time

    ModbusMaster &node;
    //ModbusConf modbusConf;
    const uint8_t cols;                                    // real LCD colums - tbc: not needed
    const uint8_t rows;                                    // real LCD rows   - tbc: not needed   
    
  public:
    LiquidCrystal_Modbus(ModbusMaster &node, uint8_t cols = 16, uint8_t rows = 2) : 
      node{node},
      //modbusConf{modbusConf},
      cols{cols},
      rows{rows}
      {}
   
/*  
    init()
    LCD API: Initializes the display.
    Clears the screen and puts the cursor to 0, 0
    noiasca: begin() methods are on implementation level
             and therefore this alias also
*/
    
    int begin(byte a = 0, byte b = 0, byte c = 0)
    {
      (void)a;  // ignore any parameter
      (void)b;
      (void)c;
      byte result = node.writeSingleRegister(addrBegin, 0);
      delay(wait);
      return result;
    }
    
    void init(byte a = 0, byte b = 0, byte c = 0) {begin(a, b, c);}
    
/* 
    setDelay(Cmd,Char)
    LCD API: Lets you override the static delays in the library.
    Some displays require delays to allow command to complete
    Typically there are two different delays.
    One for LCD commands, and one when sending characters to the display.
    The library should set the default for the display.
    noiasca: only empty alias
             no implementaion in this library
*/
    void setDelay(uint8_t, char);

/*
    write(value)
    LCD API: Raw Write a value to the display
    noiasca: in this class this write is a single byte write to the hardware interface
*/    
    size_t write(uint8_t value) {      // decided against inline
      DEBUG_PRINTLN(F("Modbus write"));
      node.writeSingleRegister(addrWrite, value);
      //byte result = node.writeSingleRegister(addrWrite, value);
      //Serial.print(F("Result ")); Serial.println(result, HEX);
      delay(wait);
      return 1;
    }
    
/*
    writeOld(value)
    noiasca: in this class this write is a single byte write to the hardware interface
             it advices the slave to use the local writeOld for special characters > 0xC0    
*/
    size_t writeOld(uint8_t value)                               // legacy write of special characters >=0xC0
    {
      DEBUG_PRINTLN(F("Modbus write"));
      node.writeSingleRegister(addrWriteOld, value);
      //byte result = node.writeSingleRegister(addrWriteOld, value);
      //Serial.print(F("Result ")); Serial.println(result, HEX);
      delay(wait);
      return 1;
    }    
    

/*
    LCD API: Send a command to the display, for commands not supported by the library
    noiasca: this a "transmit to LCD" with "Register select" = 0 Instruction Register
    (so something like a data command in contrast with 1 Data Register doesn't exist officially)
*/
    void command(uint8_t);

/*  
    LCD API: Clear the display and place cursor at 0,0
*/
    int clear()
    {
      byte result = node.writeSingleRegister(addrClear, 0);
      // slow command, therefore a long delay
      delay(1500);
      return result;
    };  
    
/*  
    LCD API: Home the cursor to 0,0 and leave displayed characters
*/
    int home()
    {
      byte result = node.writeSingleRegister(addrHome, 0);
      // slow command, therefore a long delay
      delay(1500);
      return result;
    }

/*
    setCursor(Row, Col)
    LCD API: Where Row 0-MAXLINEs, and Col 0-MAXCOLUMNS
    every common LCD implmentation is column - row, therefore also in this library
*/
    int setCursor(byte col, byte row)
    {
      uint16_t out = ((uint16_t)col) << 8 | row;
      byte result = node.writeSingleRegister(addrSetCursor, out);
      delay(wait);
      return result;
    }
    
/*
    cursor_on()
    LCD API: Turn the block cursor on
    noiasca: keep as alias only
*/
    int cursor()
    {
      byte result = node.writeSingleRegister(addrCursor, 255);
      delay(wait);
      return result;
    }
    void cursor_on() {cursor();};
    
/* 
    cursor_off()
    LCD API: Turn the block cursor off
    noiasca: keep as alias only
*/
    int noCursor()
    {
      byte result = node.writeSingleRegister(addrCursor, 0);
      delay(wait);
      return result;
    }
    
    void cursor_off() {noCursor();};

/*
    blink_on()
    LCD API: Turn on the blinking underline cursor
    noiasca: keep as alias only - use blink() instead
*/
    int blink()
    {
      byte result = node.writeSingleRegister(addrBlink, 255);
      delay(wait);
      return result;
    }

    void blink_on() {blink();};
    
/*  
    blink_off()
    LCD API: Turn off the blinking underline cursor 
    noiasca: keep as alias only - use noBlink() instead
*/
    int noBlink()
    {
      byte result = node.writeSingleRegister(addrBlink, 0);
      delay(wait);
      return result;
    }
    void blink_off() {noBlink();};

// Optional Functions:
// Should be included in the library, but might not be functional 

/*
    setBacklight(val)
    LCD API optional: Set the backlight off/on, or set Backlight brightness (0-255)
    If the display only has the option to turn the backlight on and off: 0 = off, >0 = on
    noiasca: available but, implemented in the HW classes
*/

    int backlight(void) {
      return setBacklight(255);
    }
    
    int noBacklight(void) {
      return setBacklight(0);
    }
    
    int setBacklight(uint8_t newVal) {
      byte result = node.writeSingleRegister(addrSetBacklight, newVal);
      delay(wait);
      return result;
    }
  
/* 
    setContrast(val)
    LCD API optional: Set the contrast value of the display (0-255)
    noiasca: empty alias only
*/
    void setContrast(int);

/* 
    on()
    LCD API optional: Turn the LCD display on
    If the display does not have an option to turn on/off just turn backlight on/off
*/
    void on();

/* 
    off()
    LCD API optional: Turn the LCD display off
*/
    void off();

/*
    status()
    LCD API optional: Return the status of the display
    Returns the FIFO buffer on the robot-electronics display
    Can be used to get r/w status of hardwired displays 
*/

// Extend Functions:
// Things that are specific to a particular display 

/*
    load_custom_character(Char_Num,*Rows)
    LCD API extend: Load up a custom character [0-7]
    *Rows is an array of 7 bytes that contain bitmap
    To send custom character to display use lcd.write(0-7);
    See sample sketch on how to implement this.
    noiasca: most libraries use "createChar" instead
             so the LCD API name is used as alias only
             this library expects a 5x8(!) character - not 7 bytes like the API description
*/

    void createChar(uint8_t address, const uint8_t charmap[])       // all major libraries use this method name 
    {
      // the buffer contains 8 bitpatterns and in the last field the destination character
      for (size_t i = 0; i < 8; i++)
      {
        node.setTransmitBuffer(i, charmap[i]);  // unpacked: one uint16_t - one byte.
      }
      node.setTransmitBuffer(8, address);       // the ninth byte transmits the destination character
      // transmit
      node.writeMultipleRegisters(addrCreateChar, 9);  // destination Register, HowManyBytes      
      delay(wait);
    }    
    
    void load_custom_character(const uint8_t Char_Num , const uint8_t Rows[]) {createChar(Char_Num, Rows);};
    //void createChar_P(const uint8_t, const uint8_t *);     // the implementation for custom characters in PROGMEM      
   
/*
    keypad()
    LCD API extend: Read value from keypad
    Keypad page
    noiasca: no implementation
*/

/*
    printstr(Str)
    LCD API extend: Print a string without delay. 
    noiasca: no implementation
*/

// additional methods (not from LCD API)
/*
    creates 3 German Umlauts using Special Characters 5, 6, and 7
*/
    void createUml();
    
/*
  sets the used font
  some OLED displays support extended commands
  which is used for this command
*/    
    void setFont(uint8_t value = 0);
    
/*
  there are some legacy methods which are included for backward compatibilty
*/
//#include <legacy.h>  // can't be used for the Modbus Master

/*
  some specifics for the Modbus implementation
*/

    void setDelayPostWrite(uint8_t value = 5)
    {
      wait = value;
    }
    
/*    
    send collected data as one Stream to Modbus
    currently each byte is sent in a uint16_t (addrStreamExpand)
*/
    void flush()
    {
      //Serial.println(F("flush data in one go"));
      // übergeben
      for (size_t i = 0; i < getBufferLength(); i++)
      {
        node.setTransmitBuffer(i, internal[i]);  // unpacked: one uint16_t - one byte.
      }
      // transmit
      node.writeMultipleRegisters(addrStreamExpand, getBufferLength());  // destination Register, HowManyBytes      
      // delete internal buffer of bufferPrint - to be precise, set index back to 0
      resetBuffer();
      delay(wait);
    }

};
